# -*- coding: utf-8 -*-
import os
import logging
from configparser import ConfigParser
import getpass


CONFIG_TEMPLATE = """
# Define default configuration for environment, including any common/standard
# treatments, enrollments, measures, and attributes.

# ----------------------- BEGIN TEMPLATE -----------------------

[core]
abracadabra_home={ABRACADABRA_HOME}

# Logging level
logging_level=INFO

# Default experiment values, used to reduce parameter
# footprint for common Experiment instantiations
[experiment]
control=control
treatment=treatment
enrollment=enrollment
measures=metric
attributes=attr_0,attr_1

[constants]:
default_alpha=.05
min_obs_for_z=30

[stan]:
model_cache={ABRACADABRA_HOME}/compiled_stan_models/
default_bayesian_inference_method=sample
"""

TEMPLATE_BEGIN_PATTERN = (
    '# ----------------------- BEGIN TEMPLATE -----------------------')


def expand_env_var(env_var):
    """
    Expands (potentially nested) env vars by repeatedly applying
    `expandvars` and `expanduser` until interpolation stops having
    any effect.
    """
    if not env_var:
        return env_var
    while True:
        interpolated = os.path.expanduser(os.path.expandvars(str(env_var)))
        if interpolated == env_var:
            return interpolated
        else:
            env_var = interpolated


def render_config_template(template):
    """
    Generates a configuration from the provided template + variables defined in
    current scope
    :param template: a config content templated with {{variables}}
    """
    all_vars = {k: v for d in [globals(), locals()] for k, v in d.items()}
    return template.format(**all_vars)


# ConfigParser is "old-style" class which is a classobj, not a type.
# We thus use multiple inheritance with object to fix
class AbracadabraConfigParser(ConfigParser, object):
    """
    Custom config parser, with some validations
    """
    def __init__(self, *args, **kwargs):
        super(AbracadabraConfigParser, self).__init__(*args, **kwargs)
        self.is_validated = False

    def _validate(self):
        self.is_validated = True

    def read(self, filenames):
        ConfigParser.read(self, filenames)
        self._validate()


def mk_dir(dirname):
    if not os.path.isdir(dirname):
        try:
            os.makedirs(dirname)
        except OSError as e:
            raise Exception('Could not create directory {}:\n{}'.format(dirname, e))


# Home directory and configuration locations.
# We default to ~/abracadabra and ~/abracadabra/abracadabra.cfg if not provided
if 'ABRACADABRA_HOME' not in os.environ:
    ABRACADABRA_HOME = expand_env_var('~/abracadabra')
    os.environ['ABRACADABRA_HOME'] = ABRACADABRA_HOME
else:
    ABRACADABRA_HOME = expand_env_var(os.environ['ABRACADABRA_HOME'])

mk_dir(ABRACADABRA_HOME)

if 'ABRACADABRA_CONFIG' not in os.environ:
    if os.path.isfile(expand_env_var('~/abracadabra.cfg')):
        ABRACADABRA_CONFIG = expand_env_var('~/abracadabra.cfg')
    else:
        ABRACADABRA_CONFIG = ABRACADABRA_HOME + '/abracadabra.cfg'
else:
    ABRACADABRA_CONFIG = expand_env_var(os.environ['ABRACADABRA_CONFIG'])


if 'ABRACADABRA_USER' not in os.environ:
    ABRACADABRA_USER = getpass.getuser()
    os.environ['ABRACADABRA_USER'] = ABRACADABRA_USER
else:
    ABRACADABRA_USER = os.environ['ABRACADABRA_USER']

# Write the config file, if needed
if not os.path.isfile(ABRACADABRA_CONFIG):
    logging.info(f'Creating new Abracadabra config file in: {ABRACADABRA_CONFIG}')
    with open(ABRACADABRA_CONFIG, 'w') as f:
        cfg = render_config_template(CONFIG_TEMPLATE)
        f.write(cfg.split(TEMPLATE_BEGIN_PATTERN)[-1].strip())


def coerce_value(val):
    """
    Coerce config variables to proper types
    """
    def isnumeric(val):
        try:
            float(val)
            return True
        except ValueError:
            return False

    if isnumeric(val):
        try:
            return int(val)
        except ValueError:
            return float(val)

    lower_val = str(val.lower())
    if lower_val in ('true', 'false'):
        if 'f' in lower_val:
            return False
        else:
            return True

    if ',' in val:
        return [coerce_value(v.strip()) for v in val.split(',')]
    return val


CONFIG = AbracadabraConfigParser()
CONFIG.read(ABRACADABRA_CONFIG)


# Give the entire module get/set methods
def get(section, key, **kwargs):
    """
    Retrieve typed variables from config.

    Example
    -------
    from abra import config
    # print the currently-configure ABRACADABRA_HOME directory
    print(config.get('core', 'abracadabra_home'))
    """
    return coerce_value(CONFIG.get(section, key, **kwargs))


def set(section, option, value, update=False):
    CONFIG.set(section, option, value)


def search_config(df, section, key):
    """
    Search a dataframe `df` for parameters defined in the global configuration.

    Parameters
    ----------
    df: dataframe
        raw data to search
    param_name: str
        type of parameter to search ('entities', 'metrics', or 'attributes')
    """
    available = get(section, key)
    available = [available] if not isinstance(available, list) else available
    columns = df.columns
    return [c for c in columns if c in available]


DEFAULT_ALPHA = get('constants', 'default_alpha')
MIN_OBS_FOR_Z = get('constants', 'min_obs_for_z')

STAN_MODEL_CACHE = get('stan', 'model_cache')
DEFAULT_BAYESIAN_INFERENCE_METHOD = get('stan', 'default_bayesian_inference_method')

logger = logging.getLogger(__name__)
logger.setLevel(get('core', 'logging_level'))



#!/usr/bin/python
# -*- coding: utf-8 -*-
from abra.config import DEFAULT_ALPHA
from abra.stats import Samples, RateComparison
from abra.inference.frequentist.results import FrequentistTestResults
from abra.inference import FrequentistProcedure
import numpy as np
from scipy.stats import norm


class RatesRatio(FrequentistProcedure):
    """
    Runs frequentist inference procedure to test for the difference in two sample
    rates.

    Parameters
    ----------
    hypothesis : str
        the althernative hypothesis:
        - 'larger': one-tailed test, assume variation is larger than null
        - 'smaller': one-tailed test, assume variation is smaller than null
        - 'unequal': two-tailed test, variaion mean is different than null
    var_assumptions : str
        whether to use pooled or unequal variance assumptions
        - 'pooled': assume the same variance
        - 'unequal': use Smith-Satterthwait dof when cal1ualting t-stat
    """
    def __init__(self, *args, **kwargs):
        super(RatesRatio, self).__init__(*args, **kwargs)

    def run(self, control_samples, variation_samples,
            alpha=DEFAULT_ALPHA, inference_kwargs=None):
        """
        Run the inference procedure over the samples with a selected alpha
        value.

        alpha : float in [0, 1]
            the assumed Type I error rate
        """
        if isinstance(control_samples, (list, np.ndarray)):
            control_samples = Samples(control_samples)

        if isinstance(variation_samples, (list, np.ndarray)):
            variation_samples = Samples(variation_samples)

        self.alpha = alpha
        self.comparison = RateComparison(
            samples_a=variation_samples,
            samples_b=control_samples,
            alpha=self.alpha,
            hypothesis=self.hypothesis
        )

    @property
    def stats(self):
        if not hasattr(self, '_stats'):
            self._stats = self.comparison.rates_test()
        return self._stats

    @property
    def ci(self):
        """
        Calculate confidence interval for  rates ratio. Intervals outside of 1
        support the alternative hypothesis.

        Calculation follows Li, Tang, & Wong, 2008, using the MOVER-R method,
        and Rao Score intervals (Altman et al, 2000) for individual rate interval
        estimates (aka "FRSI")

        Returns
        -------
        CIs : list
            [(CI_lower, CI_upper), (CI_lower_percentile, CI_upper_percentile)]

        References
        ----------
        Li H.Q, Tang ML, Wong WK (2008) Confidence intervals for ratio of two
            Poisson rates using the method of variance estimates recovery.
            Biometrical Journal 50 (2008)
        Altman D., Machin D., Bryant TN. et al. (2000) "Statistics with confidence"
            (2nd). BMJ Books: Bristol.
        """

        def rao_score_interval(X, z, t):
            # individual rate interval method 2 (Altman et al., 2000)
            a = X + .5 * z**2.
            b = z * np.sqrt(X + .25 * z**2.)
            return (a - b) / t, (a + b) / t

        if self.hypothesis == 'larger':
            z = norm.ppf(1 - self.alpha)
        elif self.hypothesis == 'smaller':
            z = norm.ppf(self.alpha)
        elif self.hypothesis == 'unequal':
            z = np.abs(norm.ppf(1 - self.alpha / 2.))

        control = self.comparison.d2
        variation = self.comparison.d1

        X1, t1 = control.data.sum(), control.nobs
        X2, t2 = variation.data.sum(), variation.nobs

        lam_1 = X1 / t1
        lam_2 = X2 / t2

        lam_2_lam_1 = lam_2 * lam_1

        l2, u2 = rao_score_interval(X1, z, t1)
        l1, u1 = rao_score_interval(X2, z, t2)

        # Gu et al, 2008; Eq 3
        L = (lam_2_lam_1 - np.sqrt(lam_2_lam_1 ** 2 - l1 * (2 * lam_2 - l1) * (u2 * (2 * lam_1 - u2)))) /  (u2 * (2 * lam_1 - u2))

        # Gu et al, 2008; Eq 4
        U = (lam_2_lam_1 + np.sqrt(lam_2_lam_1 ** 2 - u1 * (2 * lam_2 - u1) * (l2 * (2 * lam_1 - l2)))) / (l2 * (2 * lam_1 - l2))

        return [(L, U), self.ci_percents]

    def make_results(self):
        """
        Package up inference results
        """
        statistic_value, p_value = self.stats
        accept_hypothesis = self.accept_hypothesis(statistic_value)

        return FrequentistTestResults(
            control=self.comparison.d2,
            variation=self.comparison.d1,
            delta=self.comparison.delta,
            delta_relative=self.comparison.delta_relative,
            effect_size=self.comparison.effect_size,
            alpha=self.comparison.alpha,
            confidence_interval=self.ci,
            test_statistic=self.test_statistic,
            statistic_value=statistic_value,
            p_value=p_value,
            hypothesis=self.hypothesis_text,
            accept_hypothesis=accept_hypothesis,
            inference_procedure=self,
            power=self.comparison.power
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Function that return Stan model code for modeling data with counts or rates-generated
data.
"""


def gamma_poisson():
    """
    Return Stan model code for a Gamma-Poisson Bayesian model. The model takes
    as inputs two sequences of count variables, with each entry representing the
    total number of target events that occur over a standardized time interval.
    Must provide hyperparameters `alpha_` and `beta_` as input data fields.

    Input Data
    ----------
    {
        n_control : int             # the number of control samples
        n_variation : int           # the number of variation samples
        control : sequence[int]     # the control samples
        variation : sequence[int]   # the variation samples
        alpha_ : float              # hyperparameter
        beta_  : float              # hyperparameter
    }

    """
    return """
data {
    int<lower=0> n_control;
    int control[n_control];

    int<lower=0> n_variation;
    int variation[n_variation];

    real<lower=0> alpha_; // beta prior alpha hyperparameter
    real<lower=0> beta_; // beta prior beta hyperparameter
}

parameters {
    real<lower=0> lambda_control;
    real<lower=0> lambda_variation;
}

transformed parameters {
    real delta;
    real delta_relative;
    real effect_size;

    delta = lambda_variation - lambda_control;
    effect_size = lambda_variation / lambda_control;
    delta_relative = effect_size - 1;

}

model {
    // Gamma priors
    lambda_control ~ gamma(alpha_, beta_);
    lambda_variation ~ gamma(alpha_, beta_);

    // Poisson likelihoods
    control ~ poisson(lambda_control);
    variation ~ poisson(lambda_variation);
}

generated quantities {}
    """



#!/usr/bin/python
# -*- coding: utf-8 -*-
from abra.config import DEFAULT_ALPHA, logger
from abra.mixin import InitRepr
from statsmodels.stats.api import DescrStatsW, CompareMeans
from statsmodels.distributions.empirical_distribution import ECDF
from statsmodels.stats.power import tt_ind_solve_power, zt_ind_solve_power
from statsmodels.stats.proportion import proportions_ztest, binom_test
from scipy.stats import norm
from scipy import optimize
from pandas import DataFrame
import numpy as np


CORRECTIONS = {'b': 'bonferroni', 's': 'sidak', 'bh': 'fdr_bh'}


def bonferroni(alpha_orig, p_values):
    """
    Bonferrnoi correction.

    en.wikipedia.org/wiki/Bonferroni_correction

    Parameters
    ----------
    alpha_orig : float
        alpha value before correction
    p_values: list[float]
        p values resulting from all the tests

    Returns
    -------
    alpha_corrected: float
        new critical value (i.e. the corrected alpha)
    """
    return alpha_orig / len(p_values)


def sidak(alpha_orig, p_values):
    """
    Sidak correction.

    en.wikipedia.org/wiki/%C5%A0id%C3%A1k_correction

    Parameters
    ----------
    alpha_orig : float
        alpha value before correction
    p_values: list[float]
        p values resulting from all the tests

    Returns
    -------
    alpha_corrected: float
        new critical value (i.e. the corrected alpha)
    """
    return 1. - (1. - alpha_orig) ** (1. / len(p_values))


def fdr_bh(fdr, p_values):
    """
    Benjamini-Hochberg false-discovery rate adjustment procedure.

    pdfs.semanticscholar.org/af6e/9cd1652b40e219b45402313ec6f4b5b3d96b.pdf

    Parameters
    ----------
    fdr : float
        False Discovery Rate (q*), proportion of significant results that are
        actually false positives
    p_values: list[float]
        p values resulting from all the tests

    Returns
    -------
    alpha_corrected: float
        new critical value (i.e. the corrected alpha)
    """
    n_tests = len(p_values)

    def p_i(i):
        return i * fdr / n_tests

    p_sorted = np.sort(np.asarray(p_values))

    significant_idx = [i for i, val in enumerate(p_sorted, 1) if val <= p_i(i)]
    rank = np.max(significant_idx) if significant_idx else 1
    return p_i(rank)


def estimate_experiment_sample_sizes(
    delta,
    statistic='z',
    alpha=.05,
    power=.8,
    *args, **kwargs
):
    """
    Calculate the sample size required for each treatement in order to observe a
    difference of `delta` between control and variation groups, for a given setting
    of `alpha`, `power`.

    Parameters
    ----------
    delta : float
        The absolute difference in means between control and variation groups
    statistic : string
        Either:
            - 'z' or 't' if interpreting effect size as scaled difference of means
            - 'rates_ratio' if interpreeting effect size as the ratio of means
    alpha : float [0, 1)
            The assumed Type I error of the test
    power : float [0, 1)
        The desired statistical power of the test
    *args, **kwargs
        Model-specific arguments

    Returns
    -------
    sample_sizes : list[int]
        The estiamated sample sizes for the control and variation treatments

    Example 1: Continuous Variables
    -------------------------------
    # Estimate the sample size required to observe significant difference between
    # two binomial distributions that differ by .01 in mean probability with
    # Type I error = 0.05 (default) and Power = 0.8 (default)

    prob_control = .49
    std_control = (prob_control * (1 - prob_control))**.5  # Binomial std
    prob_variation = std_variation = .50
    delta = prob_variation - prob_control

    print(
        estimate_experiment_sample_sizes(
            delta=delta,
            statistic='z',
            std_control=std_control,
            std_variation=std_variation
        )
    )
    # [39236, 39236]

    Example 2 - Count Variables
    ---------------------------
    # Replicate Example 1 from Gu et al, 2008

    R = 4  # ratio under alternative hypothesis
    control_rate = .0005
    variation_rate = R * control_rate
    delta = variation_rate - control_rate

    print(
        estimate_experiment_sample_sizes(
            delta,
            statistic='rates_ratio',
            control_rate=control_rate,
            alpha=.05,
            power=.9,
            control_exposure_time=2.,
            sample_size_ratio=.5
        )
    )
    # [8590, 4295]
    """
    if statistic in ('t', 'z'):
        # std_control and/or std_variation are in *args, or **kwargs
        return cohens_d_sample_size(delta, alpha, power, statistic, *args, **kwargs)
    elif statistic == 'rates_ratio':
        return ratio_sample_size(alpha, power, delta, *args, **kwargs)
    else:
        raise ValueError("Unknown statistic")


def cohens_d(delta, std_control, std_variation=None):
    std_variation = std_variation if std_variation else std_control
    std_pooled = np.sqrt((std_control ** 2 + std_variation ** 2) / 2.)
    return delta / std_pooled


def cohens_d_sample_size(
    delta,
    alpha,
    power,
    statistic,
    std_control,
    std_variation=None,
    sample_size_ratio=1.
):
    """
    Calculate sample size required to observe a significantly reliable difference
    between groups a and b. Assumes Cohen's d definition of effect size and an
    enrollment ratio of 1.0 between groups a and b by default.

    Parameters
    ----------
    std_control : float
        An estiamte of the expected sample standard deviation of control
        group
    nobs_control : int
        The number of control observations.
    std_variation : float
        An estimate of the expected sample standard deviation of variation
        group. If not provided, we assume homogenous variances for the
        two groups.

    Returns
    -------
    sample_sizes : list[int]
        The estiamated sample sizes for the control and variation treatments

    Example
    -------
    # Get estimate of sample size required to observe a significant difference between
    # two binomial distributions that differ by .01 in mean probability

    prob_control = .49
    std_control = (prob_control * (1 - prob_control))**.5  # Binomial std
    prob_variation = std_variation = .50
    delta = prob_variation - prob_control

    print(
        cohens_d_sample_size(
            delta=delta,
            alpha=.05,
            power=.8,
            statistic='z',
            std_control=std_control,
            std_variation=std_variation
        )
    )
    # [39236, 39236]

    References
    ----------
    Cohen, J. (1988). Statistical power analysis for the behavioral sciences
        (2nd ed.). Hillsdale, NJ: Lawrence Earlbaum Associates.
    """
    SUPPORTED_STATISTICS = ('t', 'z')
    effect_size = cohens_d(delta, std_control, std_variation)

    if statistic in SUPPORTED_STATISTICS:
        power_func = "{}t_ind_solve_power".format(statistic)
        N1 = int(
            eval(power_func)(
                effect_size,
                alpha=alpha,
                power=power,
                ratio=sample_size_ratio
            )
        )
        N2 = int(N1 * sample_size_ratio)
        return [N1, N2]
    else:
        raise ValueError("Unknown statistic, must be either {!r}".format(SUPPORTED_STATISTICS))


def ratio_sample_size(
    alpha,
    power,
    delta,
    control_rate,
    control_exposure_time=1.,
    null_ratio=1.,
    sample_size_ratio=1.,
    exposure_time_ratio=1.
):
    """
    Calculate sample size required to observe a significantly reliable ratio of
    rates between variation and control groups. Follows power calculation outlined
    in Gu et al, 2008.

    Parameters
    ----------
    control_rate : float
        The poisson rate of the control group
    control_exposure_time : float
        The number of time units of the control exposure. Default is 1.0
    null_ratio : float
        The ratio of variation to control rates under the null hypothesis.
        Default is 1.
    sample_size_ratio : float
        The ratio of sample sizes of the variation to the control groups. Default is
        1, thus assuming equal sample sizes.
    exposure_time_ratio : float
        The ratio of the variation exposure time to the control. Default is 1.0,
        thus assuming equal exposure times

    Returns
    -------
    N1, N2 : tuple
        Sample sizes for each group

    Example
    -------
    # Replicate Example 1 from Gu et al, 2008

    R = 4  # ratio under alternative hypothesis
    control_rate = .0005
    variation_rate = R * control_rate
    delta = variation_rate - control_rate

    print(
        ratio_sample_size(
            alpha=.05,
            power=.9,
            delta=delta,
            control_rate=control_rate,
            control_exposure_time=2.,
            sample_size_ratio=.5
        )
    )
    # returns [8590, 4295], which have been validated to be more accurate than
    # the result reported in Gu et al, due to rounding precision. For details
    # see "Example 2 – Validation using Gu (2008)" section of
    # http://ncss-wpengine.netdna-ssl.com/wp-content/themes/ncss/pdf/Procedures/PASS/Tests_for_the_Ratio_of_Two_Poisson_Rates.pdf

    References
    ----------
    Gu, K., Ng, H.K.T., Tang, M.L., and Schucany, W. 2008. 'Testing the Ratio of
        Two Poisson Rates.' Biometrical Journal, 50, 2, 283-298.
    Huffman, M. 1984. 'An Improved Approximate Two-Sample Poisson Test.'
        Applied Statistics, 33, 2, 224-226.
    """

    # convert absolute difference to ratio
    alternative_ratio = float(control_rate + delta) / control_rate
    variation_exposure_time = exposure_time_ratio * control_exposure_time

    z_alpha = norm.ppf(1 - alpha)
    z_power = norm.ppf(power)

    def objective(x):
        ratio_proposed = (x[1] * variation_exposure_time) / (x[0] * control_exposure_time)
        loss = np.abs(null_ratio - (alternative_ratio / ratio_proposed))
        return loss

    def con1(x):
        """General sample size ratio constraint"""
        return (float(x[1]) / x[0]) - sample_size_ratio

    def con2(x):
        """Control sample size constraint, outlined in Gu et al, 2008, Equation 10"""

        N1, N2 = x
        d = (control_exposure_time * N1) / (variation_exposure_time * N2)
        A = 2 * (1. - np.sqrt(null_ratio / alternative_ratio))
        C = np.sqrt((null_ratio + d) / alternative_ratio)
        D = np.sqrt((alternative_ratio + d) / alternative_ratio)
        return x[0] - (((z_alpha * C + z_power * D) / A) ** 2. - (3. / 8)) / (control_exposure_time * control_rate)

    constraint1 = {'type': 'eq', 'fun': con1}
    constraint2 = {'type': 'eq', 'fun': con2}
    constraints = [constraint1, constraint2]

    results = optimize.minimize(
        objective,
        (10, 10),
        bounds=((1, None), (1, None)),
        constraints=constraints,
        method='SLSQP',
        tol=1e-10
    )

    return [int(np.ceil(n)) for n in results.x]


class MultipleComparisonCorrection(InitRepr):
    """
    Perform multiple comparison adjustment of alpha based on a sequence of
    p_values that result from two or more hypothesis tests inference procedures.

    param p_values : list[float]
        A list of p_values resulting from two or more hypothesis tests.
    method :  str
        One of the following correction methods:
            'bonferroni', 'b' : one-step Bonferroni correction
            'sidak', 's' : one-step Sidak correction
            'fdr_bh', 'bh; : Benjamini/Hochberg (non-negative)
    alpha : float in (0, 1)
        the desired probability of Type I error
    reject_nul: list[bool]
        For each probablity, whether or not to reject the null hypothsis given
        the updated values for alpha.
    """
    __ATTRS__ = ['ntests', 'method', 'alpha_orig', 'alpha_corrected']

    def __init__(self, p_values, method='sidak', alpha=DEFAULT_ALPHA):
        if method not in set(list(CORRECTIONS.keys()) + list(CORRECTIONS.values())):
            raise ValueError('Correction method {!r} not supported'.format(method))

        self.method = CORRECTIONS[method] if method in CORRECTIONS else method
        self.alpha_orig = alpha
        self.alpha_corrected = eval(self.method)(alpha, p_values)
        self.ntests = len(p_values)
        self.accept_hypothesis = [p < self.alpha_corrected for p in p_values]


class EmpiricalCdf(object):
    """
    Class that calculates the empirical cumulative distribution function for a
    set of samples. Performs some additional cacheing for performance.
    """
    def __init__(self, samples):
        self.samples = samples
        self._cdf = ECDF(samples)

    @property
    def samples_cdf(self):
        """
        Return the cdf evaluated at those samples used to calculate the cdf
        parameters.
        """
        if not hasattr(self, '_samples_cdf'):
            self._samples_cdf = self.evaluate(sorted(self.samples))
        return self._samples_cdf

    def __call__(self, values):
        return self.evaluate(values)

    def evaluate(self, values=None):
        """
        Evaluate the cdf for a sequence of values
        """
        if values is None:
            values = self.samples
        return self._cdf(values)


class Samples(DescrStatsW):
    """
    Class for holding samples and calculating various statistics on those
    samples.

    Parameters
    ----------
    samples: array-like
        the data set of sample values
    """
    def __init__(self, observations, name=None):

        self.name = name
        observations = self._valid_observations(observations)
        super(Samples, self).__init__(np.array(observations))

    def _valid_observations(self, observations):
        def valid(o):
            if o is None:
                return False
            if np.isnan(o):
                return False
            return True

        observations = list(filter(valid, observations))
        if self.name:
            name_string = "{!r}".format(self.name)
        else:
            name_string = ''
        if not observations:
            raise ValueError('All {} observations are nan or None'.format(name_string))
        else:
            return observations

    def __repr__(self):
        header = "Samples(name={!r})".format(self.name if self.name else None)
        return """{}
Summary:
𝛮  : {}
𝝁  : {:1.4f}
𝝈² : {:1.4f}""".format(header, self.nobs, self.mean, self.var)

    def permute(self):
        return np.random.choice(self.data, int(self.nobs))

    def sort(self):
        if not hasattr(self, '_sorted'):
            self._sorted = sorted(self.data)
        return self._sorted

    def percentiles(self, prct=[2.5, 25, 50, 75, 97.5]):
        return np.percentile(self.data, prct)

    @property
    def cdf(self):
        if not hasattr(self, '_cdf'):
            self._cdf = EmpiricalCdf(self.data)
        return self._cdf

    def prob_greater_than(self, values):
        """
        Return the probability of being larger than values under the emprical
        CDF
        """
        return 1.0 - self.cdf(np.asarray(values, dtype=float))

    def ci(self, alpha=.05, alternative='two-sided'):
        """
        Calculate the (1-alpha)-th confidence interval around the mean.
        Assumes Gaussian approximation.

        Returns
        -------
        ci : tuple (lo, hi)
            the (1-alpha) % confidence interval around the mean estimate.
        """
        return self.zconfint_mean(alpha, alternative)[:2]

    def std_err(self, alpha=.05, alternative='two-sided'):
        """
        Returns
        -------
        std_err : tuple (lo, hi)
            the standard error interval around the mean estimate.
        """
        _alpha = alpha / 2. if alternative == 'two-sided' else alpha
        z = norm.ppf(1 - _alpha)
        ci = z * (self.var / self.nobs) ** .5
        return self.mean - ci, self.mean + ci

    def hdi(self, alpha=.05):
        """
        Calcualte the highest central density interval that leaves `alpha`
        probability remaining.

        Parameters
        ----------
        alpha: float in (0, 1)
            1 - critical mass

        Returns
        -------
        hdi: tuple (boundary_lower, boundary_upper)
            The boundary of the highest density interval for the sample distribution
        """
        credible_mass = 1 - alpha
        try:
            _hdi = highest_density_interval(self.data, credible_mass)
            return (round(_hdi[0], 4), round(_hdi[1], 4))
        except Exception as e:
            logger.warn(e)
            return (None, None)

    def hist(self, ref_val=None, *hist_args, **hist_kwargs):
        """
        Render histogram of the samples. Plot a vertical reference line, if
        requested.

        """
        from matplotlib import pyplot as plt
        pl = plt.hist(self.data.astype(float), *hist_args, **hist_kwargs)
        if ref_val is not None:
            plt.axvline(ref_val, c='gray', linestyle='--', linewidth=2)
        return pl

    def plot_probability(self, *args, **kwargs):
        """
        Evaulate and display the sample probability function.
        """
        self.prob.plot(*args, **kwargs)


class MeanComparison(CompareMeans):
    """
    Class for comparing the means of two sample distributions, provides a number
    of helpful summary statistics about the comparison.

    Parameters
    ----------
    samples_a : Samples instance
        Group a samples
    samples_b : Samples instance
        Group b samples
    alpha : float in (0, 1)
        The assumed Type I error
    test_statistic: str
        The name of the test statistic used.
            't': for t-statistic (small sample size, N <= 30)
            'z': for z-statistic (large samples size, N > 30)
    hypothesis : str
        Defines the assumed alternative hypothesis. Can be :
            'larger'
            'smaller'
            'unequal' (i.e. two-tailed test)
    """
    def __init__(self, samples_a, samples_b,
                 alpha=DEFAULT_ALPHA,
                 test_statistic='t',
                 hypothesis='larger'):

        super(MeanComparison, self).__init__(samples_a, samples_b)

        self.alpha = alpha
        self.test_statistic = test_statistic
        self.hypothesis = hypothesis
        self.warnings = []

    @property
    def pooled_variance(self):
        return ((self.d2.nobs - 1) * self.d2.var + (self.d1.nobs - 1) * self.d1.var) / (self.d2.nobs + self.d1.nobs - 2)

    @property
    def delta(self):
        return self.d1.mean - self.d2.mean

    @property
    def delta_relative(self):
        return (self.d1.mean - self.d2.mean) / np.abs(self.d2.mean)

    @property
    def effect_size(self):
        return self.delta / np.sqrt(self.pooled_variance)

    @property
    def test_direction(self):
        return self.hypothesis if self.hypothesis != 'unequal' else 'two-sided'

    @property
    def power(self):
        """
        Statistical power (i.e. 𝜷 of the comparison)
        """
        ratio = self.d1.nobs / self.d2.nobs

        f_stat = "{}t_ind_solve_power".format(self.test_statistic)

        return eval(f_stat)(
            effect_size=self.effect_size,
            nobs1=self.d2.nobs,
            alpha=self.alpha,
            ratio=ratio,
            alternative=self.test_direction
        )


class ProportionComparison(MeanComparison):
    """
    Class for comparing the proportions of two sample distributions, provides a number
    of helpful summary statistics about the comparison. In order to use the
    z-distribution, we assume normality or proportions and thus, by proxy, adequate
    sample sizes (i.e. > 30).

    Parameters
    ----------
    samples_a : Samples instance
        Group a samples
    samples_b : Samples instance
        Group b samples
    alpha : float in (0, 1)
        The assumed Type I error
    hypothesis : str
        Defines the assumed alternative hypothesis. Can be :
            'larger'
            'smaller'
            'unequal' (i.e. two-tailed test)
    """
    def __init__(self, variance_assumption='pooled', *args, **kwargs):

        super(ProportionComparison, self).__init__(test_statistic='z', *args, **kwargs)
        nobs = min(self.d1.nobs, self.d2.nobs)

        # to use Normal approx, must have large N
        if nobs < 30:
            warning = 'Normality assumption violated, at least 30 observations required. Smallest sample size is {}'.format(nobs)
            logger.warn(warning)
            self.warnings.append(warning)

        self.variance_assumption = variance_assumption

    @property
    def pooled_variance(self):
        if self.variance_assumption == 'pooled':
            p1 = self.d1.mean
            p2 = self.d2.mean
            var1 = p1 * (1 - p1)
            var2 = p2 * (1 - p2)
            return ((self.d1.nobs - 1) * var1 + (self.d2.nobs - 1) * var2) / (self.d1.nobs + self.d2.nobs - 2)
        else:  # global variance
            p = np.mean(np.r_[self.d1.data, self.d2.data])
            return p * (1 - p)

    def ztest(self):
        prop_var = self.pooled_variance
        n_1 = self.d1.nobs
        s_1 = sum(self.d1.data)
        n_2 = self.d2.nobs
        s_2 = sum(self.d2.data)
        return proportions_ztest(
            [s_1, s_2], [n_1, n_2],
            alternative=self.test_direction,
            prop_var=prop_var
        )


class RateComparison(MeanComparison):
    """
    Class for comparing the rates of two sample distributions, provides a number
    of helpful summary statistics about the comparison. Uses the exact conditional
    test based on binomial distribution, as described in Gu et al (2008)

    Parameters
    ----------
    samples_a : Samples instance
        Group a samples
    samples_b : Samples instance
        Group b samples
    alpha : float in (0, 1)
        The assumed Type I error
    hypothesis : str
        Defines the assumed alternative hypothesis. Can be :
            'larger'
            'smaller'
            'unequal' (i.e. two-tailed test)

    References
    ----------
    Gu, Ng, Tang, Schucany 2008: Testing the Ratio of Two Poisson Rates,
    Biometrical Journal 50 (2008) 2, 2008
    """
    def __init__(self, null_ratio=1., *args, **kwargs):

        super(RateComparison, self).__init__(test_statistic='W', *args, **kwargs)
        self.null_ratio = null_ratio

    @property
    def rates_ratio(self):
        """
        Return the comparison ratio of the null rates ratio and the observed
        rates ratio.
        """
        actual_ratio = float(self.d1.sum * self.d1.nobs) / float(self.d2.sum * self.d2.nobs)
        return self.null_ratio / actual_ratio

    @property
    def delta(self):
        """
        Delta is the ratio of the variation to the control rates
        """
        return self.d1.mean / self.d2.mean

    @property
    def delta_relative(self):
        return self.delta

    def rates_test(self):
        """
        Run the rates comparison hyptothesis test. Uses the W5 statistic defined
        in Gu et al., 2008

        Returns
        -------
        W : float
            The W5 statistic from Gu et al., 2008
        p_value : float
            The p-value associated with W
        """
        X1, X2 = self.d2.sum, self.d1.sum
        t1, t2 = self.d2.nobs, self.d1.nobs
        d = float(t1) / t2
        W = 2 * (np.sqrt(X2 + (3. / 8)) - np.sqrt((self.null_ratio / d) * (X1 + (3. / 8)))) / np.sqrt(1 + (self.null_ratio / d))

        if self.hypothesis == 'larger':
            p_val = 1 - norm.cdf(W)
        elif self.hypothesis == 'smaller':
            p_val = norm.cdf(W)
        elif self.hypothesis == 'unequal':
            p_val = 1 - norm.cdf(abs(W))

        return W, p_val

    @property
    def effect_size(self):
        """
        Effect size ranges from 0-1
        """
        return 1 - self.rates_ratio

    @property
    def power(self):
        """
        Return the statistical power of the current test. Follows the calculation
        from W statistic 5 in Gu et al., 2008
        """
        N2, t2 = self.d1.sum, self.d1.nobs
        N1, t1 = self.d2.sum, self.d2.nobs

        lambda_2, lambda_1 = np.abs(self.d1.mean), np.abs(self.d2.mean)
        alternative_ratio = np.abs(lambda_2 / lambda_1)
        z = norm.ppf(1 - self.alpha)

        d = float(t1 * N1) / (t2 * N2)

        A = np.abs(2. * (1. - np.sqrt(self.null_ratio / alternative_ratio)))
        B = np.sqrt(lambda_1 * t1 * N1 + (3. / 8))
        C = np.sqrt((self.null_ratio + d) / alternative_ratio)
        D = np.sqrt((alternative_ratio + d) / alternative_ratio)
        W = (A * B - z * C) / D

        return round(norm.cdf(W), 4)


def highest_density_interval(samples, mass=.95):
    """
    Determine the bounds of the interval of width `mass` with the highest density
    under the distribution of samples.

    Parameters
    ----------
    samples: list
        The samples to compute the interval over
    mass: float (0, 1)
        The credible mass under the empricial distribution

    Returns
    -------
    hdi: tuple(float)
        The lower and upper bounds of the highest density interval
    """
    _samples = np.asarray(sorted(samples))
    n = len(_samples)

    interval_idx_inc = int(np.floor(mass * n))
    n_intervals = n - interval_idx_inc
    interval_width = _samples[interval_idx_inc:] - _samples[:n_intervals]

    if len(interval_width) == 0:
        raise ValueError('Too few elements for interval calculation')

    min_idx = np.argmin(interval_width)
    hdi_min = _samples[min_idx]
    hdi_max = _samples[min_idx + interval_idx_inc]
    return hdi_min, hdi_max


class BootstrapStatisticComparison(MeanComparison):
    """
    Class for comparing a bootstrapped test statistic for two samples. Provides
    a number of helpful summary statistics about the comparison.

    Parameters
    ----------
    samples_a : Samples instance
        Group a samples
    samples_b : Samples instance
        Group b samples
    alpha : float in (0, 1)
        The assumed Type I error
    hypothesis : str
        Defines the assumed alternative hypothesis. Can be :
            'larger'
            'smaller'
            'unequal' (i.e. two-tailed test)
    n_bootstraps : int
        The number of bootstrap samples to draw use for estimates.
    statistic_function : function
        Function that returns a scalar test statistic when provided a sequence
        of samples.

    References
    ----------
    Efron, B. (1981). "Nonparametric estimates of standard error: The jackknife, the bootstrap and other methods". Biometrika. 68 (3): 589–599
    """
    def __init__(self, n_bootstraps=1000, statistic_function=None, *args, **kwargs):
        statistic_function = statistic_function if statistic_function else np.mean
        statistic_name = statistic_function.__name__
        super(BootstrapStatisticComparison, self).__init__(
            test_statistic=f"{statistic_name}", *args, **kwargs
        )
        self.statistic_function = statistic_function
        self.n_bootstraps = n_bootstraps


    def bootstrap_test(self):
        """
        Run the sample comparison hyptothesis test. Uses the bootstrapped sample statistics

        Returns
        -------
        delta: float
            The observed difference in test statistic
        p_value : float
            The p-value associated with delta
        """
        all_samples = np.concatenate([self.d1.data, self.d2.data]).astype(float)

        d1_samples = np.random.choice(all_samples, (int(self.d1.nobs), self.n_bootstraps), replace=True)
        d1_statistics = np.apply_along_axis(self.statistic_function, axis=0, arr=d1_samples)

        d2_samples = np.random.choice(all_samples, (int(self.d2.nobs), self.n_bootstraps), replace=True)
        d2_statistics = np.apply_along_axis(self.statistic_function, axis=0, arr=d2_samples)

        control_bs_samples = np.random.choice(self.d2.data, (int(self.d2.nobs), self.n_bootstraps), replace=True)
        control_statistics = np.apply_along_axis(self.statistic_function, axis=0, arr=control_bs_samples)
        self.control_bootstrap = Samples(control_statistics, name='control')

        variation_bs_samples = np.random.choice(self.d1.data, (int(self.d1.nobs), self.n_bootstraps), replace=True)
        variation_statistics = np.apply_along_axis(self.statistic_function, axis=0, arr=variation_bs_samples)
        self.variation_bootstrap = Samples(variation_statistics, name='variation')

        # The null sampling distribution of test_statistic deltas
        self.null_dist = Samples(d2_statistics - d1_statistics, name=f'{self.test_statistic}-null')

        if self.hypothesis == 'larger':
            p_val = 1 - self.null_dist.cdf(self.delta)
        elif self.hypothesis == 'smaller':
            p_val = self.null_dist.cdf(self.delta)
        elif self.hypothesis == 'unequal':
            p_val = 1 - self.null_dist.cdf(abs(self.delta))

        return self.delta, p_val


    def confidence_interval(self, alpha=.05):
        """
        Calculate the (1-alpha)-th confidence interval around the statistic delta.
        Uses bootstrapped approximation the statistic sampling distribution.

        Returns
        -------
        ci : tuple (lo, hi)
            the (1-alpha) % confidence interval around the statistic estimate.
        """
        return  self.deltas_dist.percentiles([100 * alpha, 100 * (1-alpha)])

    @property
    def deltas_dist(self):
        if not hasattr(self, '_deltas_dist'):
            d1_samples = np.random.choice(self.d1.data, (int(self.d1.nobs), self.n_bootstraps), replace=True)
            d1_statistics = np.apply_along_axis(self.statistic_function, axis=0, arr=d1_samples)

            d2_samples = np.random.choice(self.d2.data, (int(self.d2.nobs), self.n_bootstraps), replace=True)
            d2_statistics = np.apply_along_axis(self.statistic_function, axis=0, arr=d2_samples)

            self._deltas_dist = Samples(d1_statistics - d2_statistics, name=f'{self.test_statistic}-deltas')

        return self._deltas_dist

    @property
    def delta(self):
        """
        Delta is difference in test statistics
        """
        return self.deltas_dist.mean

    @property
    def delta_relative(self):
        return self.delta / np.abs(self.statistic_function(self.d2.data))

    @property
    def power(self):
        """
        Return the statistical power of the current test. Uses
        """
        critical_value = self.null_dist.percentiles(100 * (1 - self.alpha))
        return self.deltas_dist.prob_greater_than(critical_value)


def highest_density_interval(samples, mass=.95):
    """
    Determine the bounds of the interval of width `mass` with the highest density
    under the distribution of samples.

    Parameters
    ----------
    samples: list
        The samples to compute the interval over
    mass: float (0, 1)
        The credible mass under the empricial distribution

    Returns
    -------
    hdi: tuple(float)
        The lower and upper bounds of the highest density interval
    """
    _samples = np.asarray(sorted(samples))
    n = len(_samples)

    interval_idx_inc = int(np.floor(mass * n))
    n_intervals = n - interval_idx_inc
    interval_width = _samples[interval_idx_inc:] - _samples[:n_intervals]

    if len(interval_width) == 0:
        raise ValueError('Too few elements for interval calculation')

    min_idx = np.argmin(interval_width)
    hdi_min = _samples[min_idx]
    hdi_max = _samples[min_idx + interval_idx_inc]
    return hdi_min, hdi_max



#!/usr/bin/python
# -*- coding: utf-8 -*-
from scipy import stats
from matplotlib import pyplot as plt
from pandas import DataFrame
import numpy as np
from abra.utils import dict_to_object
from abra.stats import Samples
NPTS = 100
LABEL_Y_OFFSET_FACTOR = 30.

COLORS = dict_to_object(
    {
        "blue": "#4257B2",
        "light_blue": "#A1C4FD",
        "cyan": "#3CCFCF",
        "green": "#388E34",
        "light_green": "#28CC7D",
        "dark_green": "#006060",
        "yellow": "#FFCD1F",
        "salmon": "#FF725B",
        "red": "#FB3640",
        "dark_red": "#AE2024",
        "purple": "#8842C0",
        "gray": "#687174",
        "dark_gray": "#455357",
        "light_gray": "#C0CACE",
        "brown": "#665000"
    }
)

CONTROL_COLOR = COLORS.blue
VARIATION_COLOR = COLORS.green
DIFF_COLOR = COLORS.dark_gray
RESULTS_FIGSIZE = (15, 10)


class Plottable(object):
    def __init__(self, label=None, color=None):
        self.label = label
        self.color = color


class Pdf(Plottable):
    """
    Base class for plotting probability density functions.
    """
    def __init__(self, fill=True, *args, **kwargs):
        super(Pdf, self).__init__(*args, **kwargs)
        self.fill = fill

    def density(self, xs):
        """
        Evaluate
        """
        raise NotImplementedError("Implement Me")

    def xgrid(self):
        """
        Return the default x-values for plotting
        """
        raise NotImplementedError("Implement Me")

    def ppf(self, x):
        return self.dist.ppf(x)

    def cdf(self, x):
        return self.dist.cdf(x)

    def get_series(self):
        xs = self.xgrid().flatten()
        ys = self.density(xs)
        return xs, ys

    def plot(self, **plot_args):
        xs, ys, = self.get_series()
        plt.plot(xs, ys, label=self.label, color=self.color, **plot_args)
        if self.fill:
            self.plot_area(xs, ys)

    def plot_area(self, xs=None, ys=None, color=None, alpha=.25, label=None):
        xs = self.xgrid().flatten() if xs is None else xs
        ys = self.density(xs) if ys is None else ys
        color = self.color if color is None else color
        plt.fill_between(xs, ys, color=color, alpha=alpha, label=label)

    def sample(self, size):
        return self.dist.rvs(size=size)


class KdePdf(Pdf):
    """
    Estimate the shape of a PDF using a kernel density estimate.
    """
    def __init__(self, samples, *args, **kwargs):
        super(KdePdf, self).__init__(*args, **kwargs)
        self.kde = stats.gaussian_kde(samples)
        low = min(samples)
        high = max(samples)
        self._xgrid = np.linspace(low, high, NPTS + 1)

    def density(self, xs):
        return self.kde.evaluate(xs)

    def xgrid(self):
        return self._xgrid


class Pdfs(object):
    """
    Plot a sequence of Pdf instances.
    """
    def __init__(self, pdfs):
        self.pdfs = pdfs

    def plot(self):
        # labels = []
        for p in self.pdfs:
            p.plot()
        plt.legend()


class Gaussian(Pdf):
    """
    Plot a Gaussian PDF
    """
    def __init__(self, mean=0., std=1., *args, **kwargs):
        super(Gaussian, self).__init__(*args, **kwargs)
        self.mean = mean
        self.std = std
        self.dist = stats.norm(loc=mean, scale=std)

    def density(self, xs):
        return self.dist.pdf(xs)

    def xgrid(self):
        _min = self.mean - 4 * self.std,
        _max = self.mean + 4 * self.std
        return np.linspace(_min, _max, NPTS + 1)


class Pmf(Plottable):
    """
    Base class for plotting probability mass functions.
    """
    def density(self, xs):
        raise NotImplementedError("Implement Me")

    def xgrid(self):
        """
        Return the default x-values for plotting
        """
        raise NotImplementedError("Implement Me")

    def get_series(self):
        xs = self.xgrid()
        ys = self.density(xs)
        return xs, ys

    def plot(self, plot_type='step', **plot_args):
        """
        Parameters
        ---------
        plot_type: str
            The type of plot mode to use, can one of matplotlib's default plot
            types (e.g. 'bar', 'plot', 'scatter')
        """
        xs, ys, = self.get_series()
        plotfun = getattr(plt, plot_type)
        plotfun(xs, ys, label=self.label, color=self.color, **plot_args)

    def sample(self, size):
        return self.dist.rvs(size=size)


class Binomial(Pmf):
    """
    Plot a Binomial PMF
    """
    def __init__(self, n=20, p=.5, *args, **kwargs):
        super(Binomial, self).__init__(*args, **kwargs)
        self.n = n
        self.p = p
        self.dist = stats.binom(n, p)

    def density(self, xs):
        return self.dist.pmf(xs)

    def xgrid(self):
        return np.arange(0, self.n)


class Bernoulli(Pmf):
    """
    Plot a Bernoulli PDF
    """
    def __init__(self, plot_type='bar', p=0.5, *args, **kwargs):
        super(Bernoulli, self).__init__(*args, **kwargs)
        self.plot_type = plot_type
        self.p = p
        self.dist = stats.bernoulli(p)

    def density(self, xs):
        return self.dist.pmf(xs)

    def xgrid(self):
        return np.linspace(0., 1., 2)


class Poisson(Pmf):
    """
    Plot a Binomial PMF
    """
    def __init__(self, mu=1, *args, **kwargs):
        super(Poisson, self).__init__(*args, **kwargs)
        self.mu = mu
        self.dist = stats.poisson(mu)

    def density(self, xs):
        return self.dist.pmf(xs)

    def xgrid(self):
        return np.arange(0, max([1 + self.mu * 2., 11]))


def plot_interval(
    left, right, middle,
    color=None, display_text=False,
    label=None, y=0., offset=.005, fontsize=14
):
    color = color if color else 'k'
    text_y = y + offset

    if middle in (-np.inf, np.inf) and (left in (np.inf, -np.inf) or right in (np.inf, -np.inf)):
        raise ValueError('too many interval values are inf')

    _left = middle - 4 * np.abs(right) if left in (np.inf, -np.inf) else left
    _right = middle + 4 * np.abs(left) if right in (np.inf, -np.inf) else right

    plt.plot((_left, _right), (y, y), color=color, linewidth=3, label=label)
    plt.plot(middle, y, 'o', color=color, markersize=10)

    if display_text:
        label = "{}\n({}, {})".format(round(middle, 2), round(left, 2), round(right, 2))
        plt.text(middle, text_y, label, ha='center', fontsize=fontsize, color=color)


def raise_y(ax, baseline=0):
    ylims = ax.get_ylim()
    ax.set_ylim(baseline, ylims[1])
    return ax


def lower_y(ax, baseline=None):
    ylims = ax.get_ylim()

    baseline = baseline if baseline else ylims[0] - np.abs(ylims[1]) * .05

    ax.set_ylim(baseline, ylims[1])
    return ax


def visualize_gaussian_results(results, figsize=(15, 10), outfile=None, *args, **kwargs):
    """
    Visualize the results that use Gaussian approximation.
    """
    pdf_control = Gaussian(
        mean=results.control.mean,
        std=results.control.std,
        label=results.control.name,
        color=CONTROL_COLOR
    )
    pdf_variation = Gaussian(
        mean=results.variation.mean,
        std=results.variation.std,
        label=results.variation.name,
        color=VARIATION_COLOR
    )
    pdfs = Pdfs([pdf_control, pdf_variation])

    mean_diff = results.variation.mean - results.control.mean
    std_diff = ((results.control.var / results.control.nobs) + \
                (results.variation.var / results.control.nobs)) ** .5
    pdf_diff = Gaussian(mean_diff, std_diff, label='Difference', color=DIFF_COLOR)

    fig, axs = plt.subplots(3, 1, figsize=figsize)
    plt.sca(axs[0])
    pdfs.plot()
    raise_y(axs[0])
    plt.gca().get_yaxis().set_ticks([])
    plt.title("Sample Comparison")
    x_min, x_max = plt.xlim()

    plt.sca(axs[1])
    y_min, y_max = plt.ylim()
    y_dist = (y_max - y_min) / LABEL_Y_OFFSET_FACTOR
    plot_interval(
        *results.control.std_err(),
        middle=results.control.mean,
        y=y_dist,
        offset=-.015,
        color=CONTROL_COLOR,
        display_text=True,
        label=results.control.name
    )
    plot_interval(
        *results.variation.std_err(),
        middle=results.variation.mean,
        y=-y_dist,
        offset=0.005,
        color=VARIATION_COLOR,
        display_text=True,
        label=results.variation.name
    )
    plt.legend()
    plt.xlim(x_min, x_max)
    plt.gca().get_yaxis().set_ticks([])
    plt.title("Mean +/- Standard Error")

    # plot differences distribution
    plt.sca(axs[2])
    plt.axvline(0., color=DIFF_COLOR, linestyle='--', linewidth=1.5)

    if results.inference_procedure.hypothesis == 'larger':
        left_bound = results.ci[0][0]
        right_bound = np.inf
    elif results.inference_procedure.hypothesis == 'smaller':
        right_bound = results.ci[0][1]
        left_bound = np.inf
    else:
        left_bound = results.ci[0][0]
        right_bound = results.ci[0][1]

    plot_interval(
        left_bound,
        right_bound,
        mean_diff,
        color=DIFF_COLOR,
        display_text=True
    )
    plt.gca().get_yaxis().set_ticks([])
    plt.title(results.comparison_type)
    if outfile:
        plt.savefig(
            outfile,
            bbox_inches='tight',
            dpi=300
        )


def visualize_binomial_results(results, figsize=(15, 10), outfile=None, *args, **kwargs):
    """
    Visualize the results that use Gaussian approximation.
    """
    tol = 1e-4

    pmf_control = Binomial(
        p=results.control.mean,
        n=results.control.nobs,
        label=results.control.name,
        color=CONTROL_COLOR
    )

    pmf_variation = Binomial(
        p=results.variation.mean,
        n=results.variation.nobs,
        label=results.variation.name,
        color=VARIATION_COLOR
    )

    xy_control = zip(pmf_control.xgrid(), pmf_control.density(pmf_control.xgrid()))
    xy_variation = zip(pmf_variation.xgrid(), pmf_variation.density(pmf_variation.xgrid()))

    valid_xy_control = sorted([x for x in xy_control if x[1] >= tol], key=lambda x: x[0])
    valid_xy_variation = sorted([x for x in xy_variation if x[1] >= tol], key=lambda x: x[0])

    x_min = int(min(valid_xy_control[0][0], valid_xy_variation[0][0]))
    x_max = int(max(valid_xy_control[-1][0], valid_xy_variation[-1][0]))

    mean_diff = results.variation.mean - results.control.mean
    std_diff = (results.control.var / results.control.nobs + \
                results.variation.var / results.control.nobs) ** .5
    pdf_diff = Gaussian(mean_diff, std_diff, label='Difference', color=DIFF_COLOR)

    fig, axs = plt.subplots(3, 1, figsize=figsize)
    plt.sca(axs[0])

    # make plotting more scalable
    if pmf_control.n > 1000 or pmf_variation.n > 1000:
        plot_type = 'step'
    else:
        plot_type = 'bar'

    pmf_control.plot(plot_type=plot_type, alpha=.5)
    pmf_variation.plot(plot_type=plot_type, alpha=.5)
    raise_y(axs[0])
    plt.xlim(x_min, x_max)
    # plt.gca().get_xaxis().set_ticks([])
    # plt.gca().get_yaxis().set_ticks([])
    plt.legend()
    plt.title("Sample Comparison")

    plt.sca(axs[1])
    y_min, y_max = plt.ylim()
    y_dist = (y_max - y_min) / LABEL_Y_OFFSET_FACTOR
    plot_interval(
        *results.control.std_err(),
        middle=results.control.mean,
        y=y_dist,
        offset=-0.015,
        color=CONTROL_COLOR,
        display_text=True,
        label=results.control.name
    )
    plot_interval(
        *results.variation.std_err(),
        middle=results.variation.mean,
        y=-y_dist,
        offset=0.005,
        color=VARIATION_COLOR,
        display_text=True,
        label=results.variation.name
    )

    plt.legend()
    plt.gca().get_yaxis().set_ticks([])
    plt.title("Proportions +/- Standard Error")

    # Differences plot
    plt.sca(axs[2])
    plt.axvline(0., color=DIFF_COLOR, linestyle='--', linewidth=1.5)

    # xs = pdf_diff.xgrid()
    if results.inference_procedure.hypothesis == 'larger':
        left_bound = results.ci[0][0]
        right_bound = np.inf
    elif results.inference_procedure.hypothesis == 'smaller':
        right_bound = results.ci[0][1]
        left_bound = np.inf
    else:
        left_bound = results.ci[0][0]
        right_bound = results.ci[0][1]

    plot_interval(
        left_bound,
        right_bound,
        mean_diff,
        color=DIFF_COLOR,
        display_text=True
    )
    plt.gca().get_yaxis().set_ticks([])
    plt.title(results.comparison_type)
    if outfile:
        plt.savefig(
            outfile,
            bbox_inches='tight',
            dpi=300
        )


def visualize_rates_results(results, figsize=(15, 10), outfile=None, *args, **kwargs):
    fig, axs = plt.subplots(3, 1, figsize=figsize)

    # Sample Comparison plot
    plt.sca(axs[0])
    control_pmf = Poisson(
        results.control.mean,
        color=CONTROL_COLOR,
        label=results.control.name
    )

    variation_pmf = Poisson(
        results.variation.mean,
        color=VARIATION_COLOR,
        label=results.variation.name
    )
    control_pmf.plot(plot_type='bar', alpha=.5)
    variation_pmf.plot(plot_type='bar', alpha=.5)
    plt.legend()
    plt.title("Sample Comparison")

    # Rates +/- standard error plot
    plt.sca(axs[1])
    y_min, y_max = plt.ylim()
    y_dist = (y_max - y_min) / LABEL_Y_OFFSET_FACTOR
    plot_interval(
        *results.control.std_err(),
        middle=results.control.mean,
        y=y_dist,
        offset=-0.015,
        color=CONTROL_COLOR,
        display_text=True,
        label=results.control.name
    )
    plot_interval(
        *results.variation.std_err(),
        middle=results.variation.mean,
        y=-y_dist,
        offset=0.005,
        color=VARIATION_COLOR,
        display_text=True,
        label=results.variation.name
    )
    plt.legend()
    plt.gca().get_yaxis().set_ticks([])
    plt.title("Rates +/- Standard Error")

    # Differences plot
    plt.sca(axs[2])

    plot_interval(
        *results.ci[0],
        middle=results.delta,
        color=DIFF_COLOR,
        display_text=True
    )
    plt.axvline(1., color=DIFF_COLOR, linestyle='--', linewidth=1.5)
    plt.gca().get_yaxis().set_ticks([])
    plt.title(results.comparison_type)
    if outfile:
        plt.savefig(
            outfile,
            bbox_inches='tight',
            dpi=300
        )


def visualize_bootstrap_results(results, figsize=(15, 10), outfile=None, plot_type='bar', *args, **kwargs):
    fig, axs = plt.subplots(3, 1, figsize=figsize)

    # Sample Comparison plot
    plt.sca(axs[0])
    
    if plot_type == 'bar':
        bins = 50 if results.control.nobs >= 100 or results.variation.nobs >= 100 else 20
        results.control.hist(bins=bins, color=CONTROL_COLOR, alpha=.5, label=results.control.name)
        results.variation.hist(bins=bins, color=VARIATION_COLOR, alpha=.5, label=results.variation.name)
    else:
        control_pmf = KdePdf(
            samples=results.control.data,
            color=CONTROL_COLOR,
            label=results.control.name
        )

        variation_pmf = KdePdf(
            samples=results.variation.data,
            color=VARIATION_COLOR,
            label=results.variation.name
        )
        control_pmf.plot(alpha=.5)
        variation_pmf.plot(alpha=.5)

    plt.legend()
    plt.title("Sample Comparison")

    # Bootstrapped statistic +/- HDI
    plt.sca(axs[1])
    y_min, y_max = plt.ylim()
    y_dist = (y_max - y_min) / LABEL_Y_OFFSET_FACTOR
    plot_interval(
        *results.aux['control'].hdi(),
        middle=results.aux['control'].mean,
        y=y_dist,
        offset=-0.015,
        color=CONTROL_COLOR,
        display_text=True,
        label=results.control.name
    )
    plot_interval(
        *results.aux['variation'].hdi(),
        middle=results.aux['variation'].mean,
        y=-y_dist,
        offset=0.005,
        color=VARIATION_COLOR,
        display_text=True,
        label=results.variation.name
    )
    plt.legend()
    plt.gca().get_yaxis().set_ticks([])
    plt.title(f"Bootstrap({results.test_statistic}) +/- 95% HDI")

    # Differences plot
    plt.sca(axs[2])

    plot_interval(
        *results.ci[0],
        middle=results.delta,
        color=DIFF_COLOR,
        display_text=True
    )
    plt.axvline(0., color=DIFF_COLOR, linestyle='--', linewidth=1.5)
    plt.gca().get_yaxis().set_ticks([])
    plt.title(f"{results.comparison_type}({results.test_statistic})")
    if outfile:
        plt.savefig(
            outfile,
            bbox_inches='tight',
            dpi=300
        )

def visualize_bayesian_results(results, figsize=RESULTS_FIGSIZE, outfile=None, *args, **kwargs):
    fig, axs = plt.subplots(3, 1, figsize=figsize)

    def get_central_tendency_params(results):
        if 'p_control' in results.traces.variables:
            return 'p_control', 'p_variation', '$p$ (proportion)'
        elif 'mu_control' in results.traces.variables:
            return 'mu_control', 'mu_variation', '$\\mu$ (mean)'
        elif 'lambda_control' in results.traces.variables:
            return 'lambda_control', 'lambda_variation', '$\\lambda$ (rate)'
    HDI = 0.95
    HDI_PRCT = round(HDI * 100)
    plt.sca(axs[0])
    ctps = get_central_tendency_params(results)
    results.traces.plot(
        ctps[0],
        label=results.control.name,
        color=COLORS.blue,
        alpha=.4
    )
    results.traces.plot(
        ctps[1],
        label=results.variation.name,
        color=COLORS.green,
        alpha=.4,
        title='Comparison of {}'.format(ctps[2])
    )
    plt.legend()
    lower_y(axs[0])
    x_min, x_max = plt.xlim()

    plt.sca(axs[1])
    y_min, y_max = plt.ylim()
    y_dist = (y_max - y_min) / LABEL_Y_OFFSET_FACTOR
    results.traces.plot(
        ctps[0],
        label=results.control.name,
        color=COLORS.blue,
        hdi=HDI,
        include_hist=False,
        y=y_dist,
        offset=-0.015
    )

    results.traces.plot(
        ctps[1],
        label=results.variation.name,
        color=COLORS.green,
        hdi=HDI,
        include_hist=False,
        y=-y_dist,
        offset=0.005
    )
    lower_y(axs[1])
    plt.xlim([x_min, x_max])
    plt.title(f"{ctps[2]} +/- {HDI_PRCT}% HDI")

    plt.sca(axs[2])
    results.traces.plot(
        'delta',
        hdi=1 - results.alpha,
        ref_val=0.0,
        color=COLORS.dark_gray,
        title='Differences in {}'.format(ctps[2])
    )
    lower_y(axs[2])

    if outfile:
        plt.savefig(
            outfile,
            bbox_inches='tight',
            dpi=300
        )

class Traces(object):
    """
    Container class for analyzing the results of Bayesian inference procedure.

    Parameters
    ----------
    traces: dict
        Key-value pairs of parameters:samples, extracted from a Bayesian inference
        procedure.
    """

    def __init__(self, traces):
        self.variables = []
        for k, v in list(traces.items()):
            if k != "lp__":
                self.variables.append(k)
                setattr(self, k, Samples(v))
        self.summarize()

    def summarize(self):
        prct = [2.5, 25, 50, 75, 97.5]
        values = []
        columns = []
        for v in self.variables:
            trace = getattr(self, v)
            _mean = trace.mean
            _hdi = trace.hdi()
            _std = trace.std
            _percentiles = trace.percentiles(prct)
            values.append(np.r_[_mean, _hdi, _std, _percentiles])
        columns = ['mean', 'hdi_lower', 'hdi_upper', 'std'] + ["{}%".format(p) for p in prct]
        self._summary = DataFrame(values, columns=columns, index=self.variables)

    @property
    def summary(self):
        return self._summary

    def plot(
        self, variable, label=None,
        color=None, ref_val=None, alpha=.25,
        bins=None, title=None,
        hdi=None, outfile=None,
        include_hist=True,
        offset=5,
        y=0.
    ):
        """
        Plot the histogram of a variable trace

        Parameters
        ----------
        variable : str
            The name of one of self.variables to plot
        label : str
            Alternative label for the legend
        ref_val : float
            A reference value location at which to draw a vertical line
        alpha : float in [0 1)
            The transparency of the histogram. Ignored if `include_hist=False`
        bins : int
            The number of histogram bins. Ignored if `include_hist=False`
        title : str
            The title of the plot
        hdi : float in [0, 1]
            The amount of probability mass within the Highest Density Interval
            to display on the histogram.
        y : float
            The y offset for interval plots. Ignored if `hdi is None`
        offset : float
            The text offset for interval plots. Ignored if `hdi is None`
        outfile : str
            The name of an output file to save the figure to.
        """
        from matplotlib import pyplot as plt  # lazy import
        from abra.vis import plot_interval

        if (include_hist is False) and (hdi is None):
            raise ValueError('include_hist must be True if hdi is None')

        if variable not in self.variables:
            raise ValueError('Variable `{}` not available'.format(variable))

        label = label if label else variable
        trace = getattr(self, variable)
        
        if include_hist:
            if bins is None:
                bins = int(len(trace.data) / 50.)
            trace.hist(
                color=color,
                alpha=alpha,
                bins=bins,
                ref_val=ref_val,
                label=label
            )
        
        if hdi is not None:  # highest density interval
            median = round(trace.percentiles(50), 3)
            _hdi = [round(h, 3) for h in trace.hdi(1 - hdi)]
            plot_interval(
                *_hdi,
                middle=median,
                display_text=True,
                offset=offset,
                y=y,
                color=color
            )

        if title is None:
            if ref_val is not None:
                gt = round(100 * trace.prob_greater_than(ref_val))
                title = " {}% < {} = {} < {}%".format(100 - gt, variable, ref_val, gt)
            else:
                title = ''
        plt.title(title, fontsize=14)

        if outfile:
            plt.savefig(
                outfile,
                bbox_inches='tight',
                dpi=300
            )



#!/usr/bin/python
# -*- coding: utf-8 -*-
from datetime import datetime
from abra.dataset import Dataset
from abra.stats import Samples, MultipleComparisonCorrection
from abra.mixin import InitRepr
from abra.config import DEFAULT_ALPHA
from abra.hypothesis_test import HypothesisTestSuiteResults
import copy


class Experiment(InitRepr):
    """
    Parameters
    ----------
    data : DataFrame
        the tabular data to analyze, must have columns that correspond
        with `treatment`, `measures`, `attributes`, and `enrollment` if any
        of those are defined.
    treatment : str
        The column in `data` that identifies the association of each
        enrollment in the experiment with one of the experiment conditions. If
        None provided, the global config is searched to identify potential
        treatments in `data`.
    measures : list[str]
        Columns in the dataset that are associated with indicator measurements.
        If None provided, the global config is searched to identify potential
        measures in `data`.
    attributes : list[str]
        the columns in `data` that define segmenting attributes
        associated with each enrollment in the experiment. If None provided, the
        global config is searched to identify potential attributes in `data`.
    name : str
        the name of the experiment
    date : datetime.datetime
        the datetime of the experiment analysis
    enrollment_strategy : str
        metadata for the type of enrollment used to generate the data
    meta : list[str]
        Any additional columns in `data` that should be included in the experiment
        dataset. These columns can be used for additional / custom segmentations.
    """
    __ATTRS__ = ["name", "enrollment_strategy", "date"]

    def __init__(
        self,
        data,
        treatment=None,
        measures=None,
        attributes=None,
        date=None,
        name=None,
        enrollment_strategy=None,
        meta=None
    ):

        # metadata
        if date is None:
            self.date = datetime.now().date()
        elif isinstance(date, datetime):
            self.date = date
        else:
            raise ValueError('date parameter must be a datetime.datetime object')

        self.name = name
        self.enrollment_strategy = enrollment_strategy

        self.dataset = Dataset(data, treatment, measures, attributes, meta)

    @property
    def ds(self):
        """
        Shorthand for the experiment dataset instance
        """
        return self.dataset

    @property
    def measures(self):
        return self.dataset.measures

    @property
    def attributes(self):
        return self.dataset.attributes

    @property
    def observations(self):
        return self.ds.data

    def run_test(
        self,
        test,
        alpha=DEFAULT_ALPHA,
        correction_method=None,
        display_results=False,
        visualize_results=False,
        inference_kwargs=None
    ):
        """
        Given a HypothesisTest, run the test and return the results.

        Parameters
        ----------
        test : HypothesisTest
            A hypothesis test object
        alpha : float
            The Type I error assumed by the experimenter when running the test
        correction_method: str
            Correction method used when running test (if any)
        display_results : boolean
            Whether to print the test results to stdout
        visualize_results : boolean
            Whether to render a visual representation of the results

        Returns
        -------
        test_results: an instance of a HypothesisTestResults or sublass
            The results of the statistical test.
        """

        control_obs = test.filter_variations(self.dataset, test.control, self.ds.treatment)
        control_obs = test.filter_segments(control_obs)
        control_obs = test.filter_metrics(control_obs)
        control_samples = Samples(control_obs, name=test.control)

        variation_obs = test.filter_variations(self.dataset, test.variation, self.ds.treatment)
        variation_obs = test.filter_segments(variation_obs)
        variation_obs = test.filter_metrics(variation_obs)
        variation_samples = Samples(variation_obs, name=test.variation)

        test_results = test.run(
            control_samples,
            variation_samples,
            alpha=alpha, inference_kwargs=inference_kwargs
        )
        test_results.correction_method = correction_method

        if display_results:
            test_results.display()

        if visualize_results:
            test_results.visualize()

        return test_results

    def run_test_suite(
        self,
        test_suite,
        alpha=DEFAULT_ALPHA,
        display_results=False,
        visualize_results=False,
        inference_kwargs=None
    ):
        """
        Given a HypothesisTestSuite, run all tests, perform multiple comparison
        alpha correction, and adjust inference.

        ----------
        test : HypothesisTest
            A hypothesis test object
        alpha : float
            The Type I error assumed by the experimenter when running the test
        display_results : boolean
            Whether to print the test results to stdout
        visualize_results : boolean
            Whether to render a visual representation of the results

        Returns
        -------
        test_results: an instance of a HypothesisTestSuiteResults or subclass
            The results of the statistical test suite.
        """
        corrected_tests = [copy.deepcopy(t) for t in test_suite.tests]

        # run original tests
        original_test_results = [
            self.run_test(test, alpha, inference_kwargs=inference_kwargs)
            for test in test_suite.tests
        ]

        # get p_values for multiple comparison procedure
        p_values = [t.p_value for t in original_test_results]

        correction_method = test_suite.correction_method
        correction = MultipleComparisonCorrection(
            p_values=p_values,
            alpha=alpha,
            method=correction_method
        )

        # rerun tests with updated alpha
        corrected_test_results = []
        for test in corrected_tests:
            corrected_result = self.run_test(
                test,
                alpha=correction.alpha_corrected,
                correction_method=correction.method
            )
            corrected_result.correction_method = correction_method
            corrected_result.render_tables()  # update display params
            corrected_test_results.append(corrected_result)

        return HypothesisTestSuiteResults(
            test_suite.tests,
            original_test_results,
            corrected_test_results,
            correction
        )



"""
Each function returns Stan model code that are appropriate for modeling data with
continuous, symmetric distributions.
"""


def gaussian():
    """
    Return Stan code for Gaussian model with Uniform priors on variance
    and Gaussian priors on the mean. Must provide hyperparameter `std_max` as
    input data field.

    Input Data
    -----------
    {
        n_control : int               # the number of control samples
        n_variation : int             # the number of variation samples
        control : sequence[float]     # the control samples
        variation : sequence[float]   # the variation samples
        std_max : float               # hyperparameter
    }
    """
    return """
data {
    int<lower=0> n_control;
    int<lower=0> n_variation;

    real control[n_control];
    real variation[n_variation];

    real<lower=0> std_max;
}

parameters {
    real mu_control;
    real mu_variation;

    real<lower=0, upper=std_max> sigma_control;
    real<lower=0, upper=std_max> sigma_variation;
}

transformed parameters {
    real delta;
    real delta_relative;
    real effect_size;

    real mean_control;
    real mean_variation;

    real<lower=0, upper=std_max> std_control;
    real<lower=0, upper=std_max> std_variation;

    mean_control = mean(control);
    mean_variation = mean(variation);

    std_control = sd(control);
    std_variation = sd(variation);

    delta = mu_variation - mu_control;
    delta_relative = (mu_variation / mu_control) - 1.0;
    effect_size = delta / sqrt((sigma_control^2 + sigma_variation^2) / 2);
}

model {
    // Gaussian priors on means
    mu_control ~ normal(mean_control, std_control);
    mu_variation ~ normal(mean_variation, std_variation);

    // Uniform priors on standard deviations
    sigma_control ~ uniform(0, std_max);
    sigma_variation ~ uniform(0, std_max);

    // Gaussian Likelihood
    control ~ normal(mu_control, sigma_control);
    variation ~ normal(mu_variation, sigma_variation);
}
generated quantities {}
    """


def exp_student_t():
    """
    Return Stan model code for a Student's t Bayesian model with an Exponential
    prior on the degrees of freedom, Gaussian priors on the location, and positive
    Uniform prior on the scale inverse precision of the distribution. Must provide
    hyperparameters `precision` and `std_max` as input data fields.

    The Student's t model will generally be more robust to large outliers than
    the Gaussian model

    Input Data
    ----------
    {
        n_control : int               # the number of control samples
        n_variation : int             # the number of variation samples
        control : sequence[float]     # the control samples
        variation : sequence[float]   # the variation samples
        mean_control : float          # the sample mean for control samples
        mean_variation : float        # sample mean for variation samples
        std_control : float           # the control sample std dev
        std_variation : float         # the variation sample std dev
        precision : float             # hyperparameter
        std_max : float               # hyperparameter
    }

    """
    return """
data {
    int<lower=0> n_control;
    real control[n_control];

    int<lower=0> n_variation;
    real variation[n_variation];

    real<lower=0> std_max;
    real<lower=1> precision;
}

parameters {
    real mu_control;
    real<lower=0> sigma_control;

    real mu_variation;
    real<lower=0> sigma_variation;

    real eta;
}

transformed parameters {
    real mean_control;
    real mean_variation;

    real<lower=0, upper=std_max> std_control;
    real<lower=0, upper=std_max> std_variation;

    real lambda_control;
    real lambda_variation;
    real delta;
    real delta_relative;
    real effect_size;

    mean_control = mean(control);
    mean_variation = mean(variation);

    std_control = sd(control);
    std_variation = sd(variation);

    // student t requires precisions
    lambda_control = 1 / sigma_control^2;
    lambda_variation = 1 / sigma_variation^2;

    delta = mu_variation - mu_control;
    delta_relative = (mu_variation / mu_control) - 1.0;
    effect_size = delta / sqrt((sigma_control^2 + sigma_variation^2) / 2);
}

model {
    // Gaussian priors on mean
    mu_control ~ normal(mean_control, std_control);
    mu_variation ~ normal(mean_variation, std_variation);

    // Uniform prior on variance
    sigma_control ~ uniform(0, std_max);
    sigma_variation ~ uniform(0, std_max);

    // Exponential prior on student's degrees of freedom
    eta ~ exponential(1 / precision);

    // Student-t likelihoods
    control ~ student_t(eta, mu_control, lambda_control);
    variation ~ student_t(eta, mu_variation, lambda_variation);
}
generated quantities {}
    """



#!/usr/bin/python
# -*- coding: utf-8 -*-
import os
import numpy as np
import pandas as pd
import string
from pandas import DataFrame


def dict_to_object(item):
    """
    Recursively convert a dictionary to an object.
    """
    def convert(item):
        if isinstance(item, dict):
            return type('DictToObject', (), {k: convert(v) for k, v in item.items()})
        if isinstance(item, list):
            def yield_convert(item):
                for index, value in enumerate(item):
                    yield convert(value)
            return list(yield_convert(item))
        else:
            return item
    return convert(item)


def ensure_dataframe(data, data_attr='data'):
    """
    Check if an object is a dataframe, and if not, check if it has an
    attribute that is a dataframe.
    """
    if not isinstance(data, DataFrame):
        if hasattr(data, data_attr) and isinstance(getattr(data, data_attr), DataFrame):
            data = getattr(data, data_attr)
        else:
            raise ValueError('`data` is incorrect format, must be a DataFrame')

    return data


def set_backend():
    """
    Set supported matplotlb backend depending on platform.
    """
    from sys import platform
    import matplotlib as mpl
    backend = 'pdf' if platform == 'darwin' else 'agg'
    mpl.use(backend)
    return backend


def generate_fake_observations(
    n_observations=10000,
    n_treatments=2,
    n_attributes=2,
    distribution='bernoulli',
    seed=123
):
    """
    Create a dataframe of artificial observations to be used for testing and demos.
    Treatments have different means, but segments defined by the attributes have
    no effect.

    Parameters
    -----------
    n_observations: int
        number of unique observations (e.g. users)
    n_treatments: int
        the number of simulated treatments. Will create an equivalent
        number of cohorts. Each successive treatment, by definition will
        have a systematically larger mean metric values (see metric
        distribution below). Note: the maximum number of treatments allowed
        is 6.
    n_attributes: int
        the number of attribute columns simulated. The number of distinct
        values taken by each attribute column is sampled uniformly between
        1 and 4. The segments defined by n_attributes have no effect on the
        sample means.
    distribution: str
        the type of metric distributions simulated
            - 'bernoulli': the mean increases from .5 by .1 for each successive treatment
            - 'gaussian': the mean increases from 0 by 1 for each successive treatment
            - 'poisson': the mean increases from 0 by 10 for each successive treatment
    seed: int
        The random number generator seed.
    """
    np.random.seed(seed)

    letters = string.ascii_uppercase
    n_treatments = min(n_treatments, 6)

    data = pd.DataFrame()
    data['id'] = list(range(n_observations))

    # add treatments
    treatments = list(letters[:n_treatments])
    data['treatment'] = np.random.choice(treatments, size=n_observations)

    # add attributes (attributes should have no effect)
    attribute_columns = ['attr_{}'.format(i) for i in range(n_attributes)]
    for ai, attr in enumerate(attribute_columns):
        attr_vals = ['A{}{}'.format(ai, a.lower()) for a in list(letters[:np.random.randint(1, 4)])]
        data[attr] = np.random.choice(attr_vals, size=n_observations)

    # add measurements, each treatment has successively larger means
    for delta, tr in enumerate(treatments):
        tr_mask = data.treatment == tr
        n_tr = sum(tr_mask)
        if 'gauss' in distribution:
            data.loc[tr_mask, 'metric'] = delta + np.random.randn(n_tr)
        elif 'bern' in distribution:
            data.loc[tr_mask, 'metric'] = list(map(bool, np.round(.1 * delta + np.random.random(n_tr))))
        elif 'poiss' in distribution:
            data.loc[tr_mask, 'metric'] = list(map(int, np.random.poisson(1 + delta , size=n_tr) ))

    return data


class suppress_stdout_stderr(object):
    """
    A context manager for doing a "deep suppression" of stdout and stderr in
    Python, i.e. will suppress all print, even if the print originates in a
    compiled C/Fortran sub-function.

    Usage
    -----
    with suppress_stdout_stderr():
        function_with_stdout_stderr()

    Notes
    -----
    This *should* not suppress raised exceptions, since exceptions are printed
    to stderr just before a script exits, and after the context manager has
    exited.

    For details see:
    https://stackoverflow.com/questions/11130156/suppress-stdout-stderr-print-from-python-functions

    """
    def __init__(self):
        # Open a pair of null files
        self.null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.save_fds = (os.dup(1), os.dup(2))

    def __enter__(self):
        # Assign the null pointers to stdout and stderr.
        os.dup2(self.null_fds[0], 1)
        os.dup2(self.null_fds[1], 2)

    def __exit__(self, *args, **kwargs):
        # Re-assign the real stdout/stderr back to (1) and (2)
        os.dup2(self.save_fds[0], 1)
        os.dup2(self.save_fds[1], 2)

        # Close all pointers
        os.close(self.null_fds[0])
        os.close(self.null_fds[1])
        os.close(self.save_fds[0])
        os.close(self.save_fds[1])


class run_context(suppress_stdout_stderr):
    """
    Context manager with optional input to allow or suppress stdout and stderr.

    Parameters
    ----------
    suppress : bool
        Whether or not to suppress stdout and stderr
    """
    def __init__(self, suppress=True, *args, **kwargs):
        if suppress:
            super(run_context, self).__init__(*args, **kwargs)

        self.suppress = suppress

    def __enter__(self):
        if self.suppress:
            super(run_context, self).__enter__()

    def __exit__(self, *args, **kwargs):
        if self.suppress:
            super(run_context, self).__exit__()


def safe_isnan(val):
    if val is not None:
        return np.isnan(val)
    return False


def safe_cast_json(data, mapping):
    """
    Apply safe casting of common problem data types (see TYPE_MAPPING).
    """
    _apply = lambda x: _safe_cast_json(x, mapping)
    if isinstance(data, (str, bool)):
        return data
    elif isinstance(data, Mapping):
        return type(data)({k: _apply(v) for k, v in list(data.items())})
    elif isinstance(data, Sequence):
        # # additional sequence processing, no None in sequences
        # _data = [_apply(v) for v in data]
        # _data = [d if d is not None else 0 for d in _data]
        return type(data)(_apply(v) for v in data)
    else:
        return mapping.get(data, data) if not safe_isnan(data) else mapping[np.nan]



#!/usr/bin/python
# -*- coding: utf-8 -*-
from abra.config import DEFAULT_ALPHA
from abra.stats import Samples, ProportionComparison
from abra.inference.frequentist.results import FrequentistTestResults
from abra.inference import FrequentistProcedure


class ProportionsDelta(FrequentistProcedure):
    """
    Runs frequentist inference procedure to test for the difference in two sample
    proportions. Assumes normality and thus, by proxy, adequate sample sizes
    (i.e. > 30).

    Parameters
    ----------
    hypothesis : str
        the althernative hypothesis:
        - 'larger': one-tailed test, assume variation is larger than null
        - 'smaller': one-tailed test, assume variation is smaller than null
        - 'unequal': two-tailed test, variaion mean is different than null
    var_assumptions : False or float in (0, 1)
        whether to calculate variance based on sample or use control or
        another variance
    """
    def __init__(self, var_assumptions='pooled', *args, **kwargs):

        super(ProportionsDelta, self).__init__(*args, **kwargs)
        self.var_assumptions = var_assumptions

    def run(self, control_samples, variation_samples,
            alpha=DEFAULT_ALPHA, inference_kwargs=None):
        """
        Run the inference procedure over the samples with a selected alpha
        value.

        alpha : float in [0, 1]
            the assumed Type I error rate
        """

        if not isinstance(control_samples, Samples):
            control_samples = Samples(control_samples)

        if not isinstance(variation_samples, Samples):
            variation_samples = Samples(variation_samples)

        self.alpha = alpha
        self.comparison = ProportionComparison(
            samples_a=variation_samples,
            samples_b=control_samples,
            alpha=self.alpha,
            hypothesis=self.hypothesis
        )

    @property
    def stats(self):
        self._stats = self.comparison.ztest()
        return self._stats

    @property
    def ci(self):
        """
        Calculate confidence interval percentiles and values.
        """
        var_assumptions = self.var_assumptions if self.var_assumptions == "pooled" else "unequal"
        ci_vals = self.comparison.zconfint_diff(self.alpha, self.hypothesis_sm, var_assumptions)

        return [ci_vals, self.ci_percents]

    def make_results(self):
        """
        Package up inference results
        """
        statistic_value, p_value = self.stats
        accept_hypothesis = self.accept_hypothesis(statistic_value)

        return FrequentistTestResults(
            control=self.comparison.d2,
            variation=self.comparison.d1,
            delta=self.comparison.delta,
            delta_relative=self.comparison.delta_relative,
            effect_size=self.comparison.effect_size,
            alpha=self.comparison.alpha,
            power=self.comparison.power,
            confidence_interval=self.ci,
            test_statistic=self.test_statistic,
            statistic_value=statistic_value,
            p_value=p_value,
            df=None,
            hypothesis=self.hypothesis_text,
            accept_hypothesis=accept_hypothesis,
            inference_procedure=self,
            warnings=self.comparison.warnings
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-
from abra.hypothesis_test import HypothesisTestResults, PrettyTable, OrderedDict
import numpy as np


class FrequentistTestResults(HypothesisTestResults):
    """
    Class for holding, displaying, and exporting frequentist hypothesis test
    results.

    alpha : float in (0, 1)
        The "significance" level, or type I error rate for the experiment
    power : float in (0, 1)
        The statistical power of the experiment, or the probability of correctly
        rejecting the null hypothesis.
    confidence_interval : tuple
        The values and percentiles associated with the lower and upper bounds of
        the confidence interval around delta, based on `alpha`
    test_statistic : str
        The name of the test statistic used to calculate the p-value for the test
    statistic_value : float
        The value of the test statistic used to calculate the p-value
    p_value : float in (0, 1)
        The p-value, based on the test value of `statistic_value`
    df : int
        The degrees of freedom for the experiment, if applicable
    hypothesis : str
        Human-readable message for interpreting the experiment
    accept_hypothesis : boolean
        Whether or not to accept the alternative hypothesis, based on `p_value` and `alpha`
        and any correction method.
    correction_method : str
        The name of the multiple comparison correction method used, if any
    """
    def __init__(self, alpha, confidence_interval,
                 p_value, hypothesis, accept_hypothesis, power=np.nan, df=np.nan,
                 test_statistic=None, statistic_value=np.nan,
                 correction_method=None,
                 warnings=None, *args, **kwargs):

        super(FrequentistTestResults, self).__init__(*args, **kwargs)
        self.alpha = alpha
        self.power = power
        self.ci = confidence_interval
        self.test_statistic = test_statistic
        self.statistic_value = statistic_value
        self.p_value = p_value
        self.df = df
        self.hypothesis = hypothesis
        self.accept_hypothesis = accept_hypothesis
        self.correction_method = correction_method
        self.model_name = self.inference_procedure.method
        self.comparison_type = self.inference_procedure.__class__.__name__
        self.estimate_relative_confidence_interval()
        self.warnings = "\n".join(warnings) if warnings else None

    def estimate_relative_confidence_interval(self):
        """
        Estimate the confidence on the relative difference using interpolation.
        """
        self.ci_relative = [
            tuple((self.ci[0] + self.control.mean) / self.control.mean - 1.),
            self.ci[1]
        ]

    def render_stats_table(self):
        alpha_string = "alpha (corrected)" if self.correction_method else "alpha"
        tbl = PrettyTable(header=False)
        tbl.add_column(
            "",
            [
                "{}".format(self.comparison_type),
                "{} CI".format(self.comparison_type),
                "CI %-tiles",
                "{}-relative".format(self.comparison_type),
                "CI-relative",
                "Effect Size",
                alpha_string,
                "Power",
                "Inference Method",
                "Test Statistic ({!r})".format(self.test_statistic),
                "p-value",
                "Degrees of Freedom",
                "Hypothesis",
                "Accept Hypothesis",
                "MC Correction",
                "Warnings",
            ],
            align="c"
        )
        tbl.add_column(
            "",
            [
                "{:1.4f} ".format(self.delta),
                "({:1.4f}, {:1.4f})".format(self.ci[0][0], self.ci[0][1]),
                "({:1.4f}, {:1.4f})".format(self.ci[1][0], self.ci[1][1]),
                "{:2.2f} %".format(100 * self.delta_relative),
                "({:1.2f}, {:1.2f}) %".format(
                    100 * self.ci_relative[0][0], 100 * self.ci_relative[0][1]
                ),
                "{:1.4f}".format(self.effect_size),
                "{:1.4f}".format(self.alpha),
                "{:1.4f}".format(self.power),
                "{!r}".format(self.model_name),
                "{:1.4f}".format(self.statistic_value),
                "{:1.4f}".format(self.p_value),
                "{!r}".format(self.df),
                "{!r}".format(self.hypothesis),
                "{!r}".format(self.accept_hypothesis),
                "{!r}".format(self.correction_method),
                "{!r}".format(self.warnings)
            ],
            align="l"
        )
        self._stats_table = str(tbl)

    def display(self):
        print(self)

    @property
    def json(self):
        """
        Add results properties that are special to Bayesian test
        """
        _json = self._base_json
        _json.update(
            OrderedDict(
                [
                    ('test_type', ['frequentist']),
                    ('p', [self.p_value]),
                    ('p_interpretation', ["p-value"]),
                    ('delta_ci', [self.ci[0]]),
                    ('ntiles_ci', [self.ci[1]]),
                    ('delta_relative_ci', [(100 * self.ci_relative[0][0], 100 * self.ci_relative[0][1])]),
                    ('ci_interpretation', ['Confidence Interval']),
                    ('p_value', [self.p_value]),
                    ('power', [self.power]),
                    ('statistic_name', [self.test_statistic]),
                    ('statistic_value', [self.statistic_value]),
                    ('df', [self.df]),
                    ('mc_correction', [self.correction_method])
                ]
            )
        )
        return _json

    def visualize(self, figsize=None, outfile=None, *args, **kwargs):
        # lazy import
        from abra.vis import (
            visualize_gaussian_results,
            visualize_binomial_results,
            visualize_rates_results,
            visualize_bootstrap_results,
            RESULTS_FIGSIZE
        )

        figsize = figsize if figsize else RESULTS_FIGSIZE
        _model_name = self.model_name.replace(" ", '').replace("-", "").replace("_", "")
        if _model_name in ('meansdelta'):
            visualize_gaussian_results(self, figsize, outfile, *args, **kwargs)
        elif _model_name in ('proportionsdelta'):
            visualize_binomial_results(self, figsize, outfile, *args, **kwargs)
        elif _model_name in ('ratesratio'):
            visualize_rates_results(self, figsize, outfile, *args, **kwargs)
        elif _model_name in ('bootstrap'):
            visualize_bootstrap_results(self, figsize, outfile, *args, **kwargs)



from abra.utils import set_backend
from abra.experiment import Experiment
from abra.dataset import Dataset
from abra.hypothesis_test import HypothesisTest, HypothesisTestSuite, CustomMetric
from abra.stats import Samples, MultipleComparisonCorrection
from abra.inference.bayesian.delta import BayesianDelta
from abra.inference.frequentist.means import MeansDelta
from abra.inference.frequentist.proportions import ProportionsDelta
from abra.inference.frequentist.rates import RatesRatio
from abra.inference.frequentist.bootstrap import BootstrapDelta

VISUALIZATION_BACKEND = set_backend()  # set backend for any visualization support

__all__ = [
    "Experiment",
    "Dataset",
    "Samples",
    "HypothesisTest",
    "HypothesisTestSuite",
    "MultipleComparisonCorrection",
    "BayesianDelta",
    "MeansDelta",
    "ProportionsDelta",
    "RatesRatio",
    "BootstrapDelta",
    "CustomMetric",
    "VISUALIZATION_BACKEND"
]



#!/usr/bin/python
# -*- coding: utf-8 -*-
from collections.abc import Sequence, Mapping
import json
from pandas import DataFrame
import numpy as np
from abra.utils import safe_cast_json, safe_isnan

BIG = 2.**32
SMALL = -2.**32
TYPE_MAPPING = {np.inf: None, -np.inf: None, np.nan: None}  # uclear what the best choice is here


class InitRepr(object):
    """
    __repr__ returns intialization command
    """
    def __repr__(self):
        class_name = self.__class__.__name__
        class_len = len(class_name)
        attr_str = ["{}={!r}".format(attr, getattr(self, attr)) for attr in self.__ATTRS__]
        attr_str = ",\n{}".format(" " * (1 + class_len)).join(attr_str)
        return f"{class_name}({attr_str})"


class Jsonable(object):
    """
    Allows object ot be converted to a json string
    """

    def to_json(self):
        return json.dumps(self, default=lambda x: x.__dict__,
                          sort_keys=True, indent=4)

    @property
    def json(self):
        return json.loads(self.to_json())

    def __repr__(self):
        return self.to_json()


class Dataframeable(object):
    """
    Export a set of attributes tabular format. Must implement a `json` property
    method that returns a dict-like with the following structure (i.e. that
    supported by pandas for generating a single-row dataframe from JSON):

    {
        "FIELD_1_NAME": [FIELD_1_VALUE],
        "FIELD_2_NAME": [[FIELD_2_VALUE_A, FIELD_2_VALUE_B]],
        "FIELD_3_NAME": [(FIELD_3_VALUE_A, FIELD_3_VALUE_B)], ...
    }
    """
    @property
    def json(self):
        return self._json()

    def _json(self):
        raise NotImplementedError('Must implement _json method')

    def to_dataframe(self, safe_cast=False):
        """
        Export result to a dataframe. If `safe_cast` is True attempt to
        casting common problematic values (see TYPE_MAPPING); this can be useful
        for ensuring type safety, for example when uploading data to a database.
        """
        _json = self.json
        if safe_cast:
            _json = safe_cast_json(_json, TYPE_MAPPING)
        return DataFrame(_json)



#!/usr/bin/python
# -*- coding: utf-8 -*-
import numpy as np
from abra.config import DEFAULT_ALPHA
from abra.stats import Samples, BootstrapStatisticComparison
from abra.inference.frequentist.results import FrequentistTestResults
from abra.inference import FrequentistProcedure


class BootstrapDelta(FrequentistProcedure):
    """
    Runs frequentist inference procedure to test for the difference in a bootstrapped
    test statistic estimate

    Parameters
    ----------
    hypothesis : str
        the althernative hypothesis:
        - 'larger': one-tailed test, assume variation is larger than null
        - 'smaller': one-tailed test, assume variation is smaller than null
        - 'unequal': two-tailed test, variaion mean is different than null
    statistic_function : callable
        Function that returns a scalar test statistic when provided a sequence
        of samples.
    """
    def __init__(self, statistic_function=None, *args, **kwargs):
        super(BootstrapDelta, self).__init__(*args, **kwargs)
        self.statistic_function = statistic_function

    def run(
        self, control_samples, variation_samples,
        alpha=DEFAULT_ALPHA, inference_kwargs=None
    ):
        """
        Run the inference procedure over the samples with a selected alpha
        value.

        alpha : float in [0, 1]
            the assumed Type I error rate
        """
        if not isinstance(control_samples, Samples):
            control_samples = Samples(control_samples)

        if not isinstance(variation_samples, Samples):
            variation_samples = Samples(variation_samples)

        self.alpha = alpha
        self.comparison = BootstrapStatisticComparison(
            samples_a=variation_samples,
            samples_b=control_samples,
            alpha=self.alpha,
            hypothesis=self.hypothesis,
            statistic_function=self.statistic_function
        )

    @property
    def stats(self):
        self._stats = self.comparison.bootstrap_test()
        return self._stats

    @property
    def ci(self):
        """
        Calculate confidence interval around deltas with percentiles and values.
        """
        ci_vals = self.comparison.confidence_interval(self.alpha)
        return [ci_vals, self.ci_percents]

    def accept_hypothesis(self, stat_value):
        """
        Use the boostrapped estimate of teh sampling distribition to test the Null
        """
        if self.hypothesis == 'larger':
            return self.alpha > self.stat_dist.prob_greater_than(stat_value)
        elif self.hypothesis == 'smaller':
            return self.alpha > 1 - self.stat_dist.prob_greater_than(stat_value)
        elif self.hypothesis == 'unequal':
            return abs(stat_value) > abs(self.stat_dist.percentiles(100 * (1 - self.alpha/2.)))
        else:
            raise ValueError('Unknown hypothesis: {!r}'.format(self.hypothesis))

    def make_results(self):
        """
        Package up inference results
        """
        statistic_value, p_value = self.stats
        self.stat_dist = self.comparison.null_dist
        accept_hypothesis = self.accept_hypothesis(statistic_value)

        aux = {
            'control': self.comparison.control_bootstrap,
            'variation': self.comparison.variation_bootstrap
        }
        return FrequentistTestResults(
            control=self.comparison.d2,
            variation=self.comparison.d1,
            delta=self.comparison.delta,
            delta_relative=self.comparison.delta_relative,
            effect_size=self.comparison.effect_size,
            alpha=self.comparison.alpha,
            power=self.comparison.power,
            confidence_interval=self.ci,
            test_statistic=self.test_statistic,
            statistic_value=statistic_value,
            p_value=p_value,
            df=None,
            hypothesis=self.hypothesis_text,
            accept_hypothesis=accept_hypothesis,
            inference_procedure=self,
            warnings=self.comparison.warnings,
            aux=aux
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Each function returns Stan model code that are appropriate for modeling
distributions of binary or proportions data.
"""


def beta_binomial():
    """
    Return Stan code for for a Beta-Binomial model that compares two samples
    based on their summary statistics. Must provide hyperparameters `alpha_` and
    `beta_` as input data fields.

    Input Data
    ----------
    {
        n_control : int       # the number of control samples
        n_variation : int     # the number of variation samples
        s_control : int       # the number of control succeses
        s_variation : int     # the number of variations successes
        alpha_ : float        # hyperparameter
        beta_ : float         # hyperparameter
    }
    """
    return """
data {
    int<lower=0> n_control; // number of control trials
    int<lower=0> s_control; // number of control successes

    int<lower=0> n_variation; // number of variation trials
    int<lower=0> s_variation; // number of control successes

    real<lower=0> alpha_; // beta prior alpha hyperparameter
    real<lower=0> beta_; // beta prior beta hyperparameter
}

parameters {
    real<lower=0, upper=1> p_control;
    real<lower=0, upper=1> p_variation;
}

transformed parameters {
    real delta;
    real delta_relative;
    real effect_size;

    delta = p_variation - p_control;
    delta_relative = p_variation / p_control - 1.0;
    effect_size = delta / sqrt((p_control * (1 - p_control) + p_variation * (1 - p_variation)) / 2);
}

model {

    // Beta priors on succes rate
    p_control ~ beta(alpha_, beta_);
    p_variation ~ beta(alpha_, beta_);

    // Binomial likelihoods
    s_control ~ binomial(n_control, p_control);
    s_variation ~ binomial(n_variation, p_variation);
}
generated quantities {}

"""


def binomial():
    """
    Returns Stan model code for a Binomial likelihood model with a uniform prior
    """
    return beta_binomial()


def bernoulli():
    """
    Return model code for a Beta-Bernouilli model to be compiled by PyStan.
    Compares two arrays of binary data, where each entry in an array indicates
    success on an associated trial. Must provide hyperparameters `alpha_` and
    `beta_` as input data fields.

    Input Data
    ----------
    {
        n_control : int             # the number of control samples
        n_variation : int           # the number of variation samples
        control : sequence[int]     # the control samples (binary)
        variation : sequence[int]   # the variation samples (binary)
        alpha_ : float              # hyperparameter
        beta_ : float               # hyperparameter
    }
    """
    return """
data {
    int<lower=0> n_control;
    int<lower=0,upper=1> control[n_control];

    int<lower=0> n_variation;
    int<lower=0,upper=1> variation[n_variation];

    real<lower=0> alpha_; // beta prior alpha hyperparameter
    real<lower=0> beta_; // beta prior beta hyperparameter
}

parameters {
    real<lower=0, upper=1> theta_control;
    real<lower=0, upper=1> theta_variation;
}

transformed parameters {
    real delta;
    real delta_relative;
    real effect_size;

    delta = theta_variation - theta_control;
    delta_relative = theta_variation / theta_control - 1.0;
    effect_size = delta / sqrt((theta_control * (1 - theta_control) + theta_variation * (1 - theta_variation)) / 2);
}

model {
    // Beta prior
    theta_control ~ beta(alpha_, beta_);
    theta_variation ~ beta(alpha_, beta_);

    // Bernoulli Likelihoods
    control ~ bernoulli(theta_control);
    variation ~ bernoulli(theta_variation);
}
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-
from abra.hypothesis_test import HypothesisTestResults, PrettyTable, OrderedDict


class BayesianTestResults(HypothesisTestResults):
    """
    Class for storing Bayesian hypothesis test statistics.

    Parameters
    ----------
    alpha : float in (0, 1)
        The "significance" level, or type I error rate for the experiment
    traces : instance of abra.vis.Traces
        The results of a Bayesian inference procedure.
    hdi : tuple
        The values and percentiles associated with the lower and upper bounds of
        the highest density interval around delta posterior, based on `alpha`
    model_name : str
        The name of the Bayesian model used.
    hypothesis : str
        Human-readable message for interpreting the experiment
    inference_method : str
        The name of the simulation method used (e.g. 'sampling', 'variational')
    data_type : python type object
        The data type of the observations used in the inference,
    hyperparameters : dict:
        Hyperparmas that specify specify model
    """
    def __init__(self, alpha, traces, hdi, hdi_relative, prob_greater, model_name,
                 hypothesis, inference_method, data_type, warnings=None, hyperparameters=None,
                 *args, **kwargs):

        super(BayesianTestResults, self).__init__(*args, **kwargs)
        self.alpha = alpha
        self.traces = traces
        self.hypothesis = hypothesis
        self.hdi = hdi
        self.hdi_relative = hdi_relative
        self.prob_greater = prob_greater
        self.model_name = model_name
        self.hypothesis = hypothesis
        self.inference_method = inference_method
        self.data_type = data_type
        self.hyperparameters = hyperparameters
        # self.accept_hypothesis = (1 - self.prob_greater) < self.alpha   # this is VERY conservative
        self.accept_hypothesis = self.prob_greater >= 1 - self.alpha  # cutoff at 75% chance p(greater)
        self.warnings = "\n".join(warnings) if warnings else None

    def render_stats_table(self):
        tbl = PrettyTable(header=False)
        tbl.add_column(
            "",
            [
                "Delta",
                "HDI",
                "HDI %-tiles",
                "Delta-relative",
                "HDI-relative",
                "Effect Size",
                "alpha",
                "Credible Mass",
                f"p({self.variation.name} > {self.control.name})",
                "Inference Method",
                "Model Hyperarameters",
                "Inference Method",
                "Hypothesis",
                "Accept Hypothesis",
                "Warnings"
            ],
            align="c"
        )
        tbl.add_column(
            "",
            [
                "{:1.4f} ".format(self.delta),
                "({:1.4f}, {:1.4f})".format(self.hdi[0][0], self.hdi[0][1]),
                "({:1.4f}, {:1.4f})".format(self.hdi[1][0], self.hdi[1][1]),
                "{:2.2f} %".format(100 * self.delta_relative),
                "({:1.2f}, {:1.2f}) %".format(
                    100 * self.hdi_relative[0][0],
                    100 * self.hdi_relative[0][1]
                ),
                "{:1.4f}".format(self.effect_size),
                "{:1.4f}".format(self.alpha),
                "{:1.4f}".format(1 - self.alpha),
                "{:1.4f}".format(self.prob_greater),
                "{!r}".format(self.model_name),
                "{!r}".format(self.hyperparameters),
                "{!r}".format(self.inference_method),
                "{!r}".format(self.hypothesis),
                "{!r}".format(self.accept_hypothesis),
                "{!r}".format(self.warnings)
            ],
            align="l"
        )
        self._stats_table = str(tbl)

    @property
    def json(self):
        """
        Add results properties that are special to Bayesian test
        """
        _json = self._base_json
        _json.update(OrderedDict([('test_type', ['bayesian']),
                                  ('p', [self.prob_greater]),
                                  ('p_interpretation', ["p(variation > control)"]),
                                  ('delta_ci', [self.hdi[0]]),
                                  ('ntiles_ci', [self.hdi[1]]),
                                  ('delta_relative_ci', [(100 * self.hdi_relative[0][0],
                                                          100 * self.hdi_relative[0][1])]),
                                  ('ci_interpretation', ['Highest Density Interval']),

                                  ('delta_hdi', [self.hdi[0]]),
                                  ('ntiles_hdi', [self.hdi[1]]),
                                  ('delta_relative_hdi', [(100 * self.hdi_relative[0][0],
                                                           100 * self.hdi_relative[0][1])]),
                                  ('credible_mass', [(1 - self.alpha)]),
                                  ('inference_method', [self.inference_method]),
                                  ('prob_greater', [self.prob_greater])]))
        return _json

    def display(self):
        print(self)

    def visualize(self, figsize=None, outfile=None, *args, **kwargs):
        from abra.vis import visualize_bayesian_results, RESULTS_FIGSIZE
        figsize = figsize if figsize else RESULTS_FIGSIZE
        visualize_bayesian_results(self, figsize, outfile, *args, **kwargs)



#!/usr/bin/python
# -*- coding: utf-8 -*-
import numpy as np
from scipy.stats import norm


class InferenceProcedure(object):
    """
    Base class for all inference procedures. Must implement the
    following methods:
      - `run()`
      - `make_results()`
    """
    def __init__(self, method=None, stat_dist=norm, *args, **kwargs):
        self.method = method
        self.stat_dist = stat_dist

    def run(self, control_samples, variation_samples,
            alpha=None, inference_kwargs=None):
        """
        Run inference procedure and return associated results
        """
        raise NotImplementedError("Implement me")

    @property
    def results(self):
        if not hasattr(self, '_results'):
            self._results = self.make_results()
        return self._results

    def make_results(self):
        """
        Format and return a HypothesisTestResults object
        """
        raise NotImplementedError("Implement me")


class FrequentistProcedure(InferenceProcedure):

    def __init__(self, hypothesis='larger', *args, **kwargs):
        super(FrequentistProcedure, self).__init__(*args, **kwargs)
        self.hypothesis = hypothesis

    def run(self, control_samples, variation_samples,
            alpha=None, inference_kwargs=None):
        raise NotImplementedError("Implement me")

    def make_results(self):
        raise NotImplementedError("Implement me")

    @property
    def control_name(self):
        return self.comparison.d2.name

    @property
    def variation_name(self):
        return self.comparison.d1.name

    @property
    def hypothesis_sm(self):
        """
        Statsmodels-compatible hypothesis
        """
        return self.hypothesis if self.hypothesis != 'unequal' else 'two-sided'

    def accept_hypothesis(self, stat_value):
        """
        Accept the null hypothesis based on the calculated statistic and statitic
        distribution.
        """
        if self.hypothesis == 'larger':
            return stat_value > self.stat_dist.ppf(1 - self.alpha)
        elif self.hypothesis == 'smaller':
            return stat_value < self.stat_dist.ppf(self.alpha)
        elif self.hypothesis == 'unequal':
            return abs(stat_value) > self.stat_dist.ppf(1 - self.alpha / 2.)
        else:
            raise ValueError('Unknown hypothesis: {!r}'.format(self.hypothesis))

    @property
    def hypothesis_text(self):

        control_name = self.control_name if self.control_name else 'control'
        variation_name = self.variation_name if self.variation_name else 'variation'

        if self.hypothesis == 'larger':
            return "{} is larger".format(variation_name)
        elif self.hypothesis == 'smaller':
            return "{} is smaller".format(variation_name)
        elif self.hypothesis == 'unequal':
            return "{} != {}".format(variation_name, control_name)
        else:
            raise ValueError('Unknown hypothesis: {!r}'.format(self.hypothesis))

    @property
    def ci_percents(self):
        if self.hypothesis == 'larger':
            return (self.alpha, np.inf)
        elif self.hypothesis == 'smaller':
            return (-np.inf, 1 - self.alpha)
        elif self.hypothesis == 'unequal':
            return ((self.alpha / 2.), 1 - (self.alpha / 2.))
        else:
            raise ValueError('Unknown hypothesis: {!r}'.format(self.hypothesis))

    @property
    def test_statistic(self):
        return self.comparison.test_statistic



from abra.inference.inference_base import InferenceProcedure, FrequentistProcedure


def get_inference_procedure(method, **infer_params):
    _method = method.lower().replace('-', '').replace('_', '').replace(' ', '')
    if _method in ('meansdelta'):
        from abra import MeansDelta as IP

    elif _method in ('proportionsdelta'):
        from abra import ProportionsDelta as IP

    elif _method in ('ratesratio'):
        from abra import RatesRatio as IP

    elif _method in ('bootstrap'):
        from abra import BootstrapDelta as IP

    elif method in (
        'gaussian',
        'bernoulli',
        'binomial',
        'beta_binomial',
        'gamma_poisson',
        'student_t',
        'exp_student_t'
    ):

        from abra import BayesianDelta as IP
        infer_params.update({"model_name": method})
    else:
        raise ValueError('Unknown inference method {!r}'.format(method))

    return IP(method=method, **infer_params)


__all__ = [
    "InferenceProcedure",
    "FrequentistProcedure"
]



#!/usr/bin/python
# -*- coding: utf-8 -*-
from abra.mixin import InitRepr, Dataframeable
from abra.utils import ensure_dataframe, run_context
from abra.config import DEFAULT_ALPHA
from abra.inference import InferenceProcedure, get_inference_procedure
from abra.stats import CORRECTIONS
from datetime import datetime
from prettytable import PrettyTable
from collections import OrderedDict
from copy import deepcopy


class CohortFilter(object):
    """
    Filtering interface for selecting cohorts of a dataframe. Powered pandas
    `query` method.

    Parameters
    ----------
    treatment_column : str
        The column associated with various treatments
    treatment_name : str
        When applied, selects out rows with `treatment_column` == `treatment_name`
    """
    def __init__(self, treatment_column, treatment_name):
        self.treatment_column = treatment_column
        self.treatment_name = treatment_name

    def apply(self, data):
        """
        Parameters
        ----------
        data : DataFrame

        Returns
        -------
        cohort_data : DataFrame
            cohort data, filtered from `data`
        """
        return data.query(f"`{self.treatment_column}` == '{self.treatment_name}'")


class SegmentFilter(object):
    """
    Simple filtering interface for selecting subsets of a dataframe. Powered
    pandas `query` interface.
    """
    def __init__(self, segment_pattern):
        self.segment_pattern = segment_pattern

    def apply(self, data):
        if self.segment_pattern is None:
            return data
        else:
            return data.query(self.segment_pattern)


class CustomMetric(object):
    """
    Metric definition that combines one or more measure columns in dataframe.

    Parameters
    ----------
    f: function
        the function that defines an operation performed on each row of the
        dataframe, resulting in a derived value.
    """
    def __init__(self, f):
        self.f = f

    def apply(self, df):
        return df.apply(self.f, axis=1)


class HypothesisTest(InitRepr):
    """
    Executes directives for running inference a procudure to compare the central
    tendencies of two groups of random samples.

    Parameters
    ----------
    inference_method : str
        The name of the inference method used to perform the hypothesis test.
        Can be one of the following:

            Frequentist Inference:
                - 'means_delta'         Continuous
                - 'proprortions_delta'  Proportions
                - 'rates_ratio'         Counts / rates
            Bayesian Inference:
                - 'gaussian'            Continuous
                - 'exp_student_t'       Continuous
                - 'bernoulli'           Proportions / binary
                - 'beta_binomial'       Proportions
                - 'binomial'            Proportions
                - 'gamma_poisson'       Counts / rates

    metric: str or CustomMetric instance (optional)
        a performance indicator over which statistical analyses
        will be performed. Each can either be a measurment in the experiment's
        dataset, or an instance of a CustomMetric. If None provided, the
    control : str
        The name of the control treatment
    variation : str
        The name of the experimental treatment
    segmentation : str or list[str] (optional)
        Defines a list of logical filter operations that follows the conventions
        used by Panda's dataframe query api and segments the treatments into subgroups.
        If a list provided, all operations are combined using logical AND
    date : datetime (optional)
        The datetime the test was run. If None provided, we assume the current
        time.
    suppress_stan_output : boolean
        Whether or not to supress the underlying Stan code optimization stdout.
    **infer_params : dict
        Any additional parameters to be passed to the inference procud
    """
    __ATTRS__ = ['metric', 'control', 'variation', 'segmentation', 'method']

    def __init__(self, inference_method, metric=None,
                 control=None, variation=None,
                 segmentation=None, date=None,
                 suppress_stan_output=True,
                 **infer_params):

        self.metric = metric
        self.control = control
        self.variation = variation
        self.inference_method = inference_method

        self.inference_procedure = get_inference_procedure(inference_method, **infer_params)
        self.suppress_stan_output = suppress_stan_output

        if isinstance(segmentation, list):
            segmentation = " & ".join(segmentation)

        self.segmentation = segmentation

        if date is None:
            self.date = datetime.now().date()

        if isinstance(self.metric, CustomMetric):
            self.metric_column = self.metric.f.__name__
        else:
            self.metric_column = metric

    def _add_custom_metric_column(self, _data):
        data = _data.copy()
        self.metric_name = self.metric.f.__name__
        data[self.metric_name] = self.metric.apply(data)
        return data

    def filter_variations(self, data, variation_name, treatment=None):
        """
        Helper function used to pull out the data series observations for a
        variation.
        """
        data = ensure_dataframe(data)
        if isinstance(self.metric, CustomMetric):
            data = self._add_custom_metric_column(data)
        else:
            self.metric_name = self.metric

        if treatment is None:
            if hasattr(self, 'treatment'):
                treatment = self.treatment
            else:
                raise ValueError("Cant't determine the treatment column, ",
                                 "please provide as `treatment` argument")

        cohort_filter = CohortFilter(treatment, variation_name)
        return cohort_filter.apply(data)

    def filter_segments(self, data):
        """
        Helper function used to filter observations for all segments.
        """
        data = ensure_dataframe(data)
        segment_filter = SegmentFilter(self.segmentation)
        return segment_filter.apply(data)

    def filter_metrics(self, data):
        """
        Helper function used to filter out observations that have invalid metric
        values.
        """
        return data[data.notna()[self.metric_name]][self.metric_name]

    def run(self, control_samples, variation_samples, alpha=DEFAULT_ALPHA, inference_kwargs=None):
        """
        Run the statistical test inference procedure comparing two groups of
        samples.

        Parameters
        ----------
        control_samples : an instance of `Samples` class
            The samples for the control treatment group
        variation_samples : an instance of `Samples` class
            The samples for the experimental treatment group
        alpha : float in (0, 1) (optional)
            Is either:
                - the 'significance level' for frequentist tests
                - the one minus the credible mass of the posterior over differences
                  between the two groups.
            If None provided, we assume an alpha = 0.05
        inference_kwargs : dict
            Any additional keyword args to be provided to the inference procedure

        Returns
        -------
        results : subclass of `TestResultsBase`
            The container class holding the summary of the experiment

        """
        if not issubclass(self.inference_procedure.__class__, InferenceProcedure):
            raise ValueError(
                "The inference procedure used must subclass `InferenceProcedure`"
            )
        self.alpha = alpha

        # when running Bayesian inference, optionally run with context
        # manager to supress stdout during Stan optimization routines
        supress_stdout = self.inference_procedure.__class__.__name__ == 'BayesianDelta' and self.suppress_stan_output
        with run_context(supress_stdout):
            self.inference_procedure.run(control_samples, variation_samples, alpha, inference_kwargs=inference_kwargs)

        results = self.inference_procedure.make_results()
        results.metric_name = self.metric_column
        results.segmentation = self.segmentation
        return results

    def copy(self, **kwargs):
        copy = deepcopy(self)
        infer_kwargs = kwargs.get("infer_kwargs", {})
        for k, v in kwargs.items():
            if hasattr(copy, k):
                setattr(copy, k, v)

        copy.inference_procedure = get_inference_procedure(copy.inference_method, **infer_kwargs)
        return copy


class TestResultsBase(Dataframeable):

    def display(self):
        raise NotImplementedError("Implement me")

    def visualize(self, figsize=None, outfile=None, *args, **kwargs):
        raise NotImplementedError("Implement me")

    def to_csv(self, fname, delimiter=','):
        """
        Export result to delimiter-separated value file
        """
        results_df = self.to_dataframe()
        results_df.to_csv(fname, sep=delimiter, encoding='utf8', index=False)

    @property
    def json(self):
        raise NotImplementedError('Must implement json property')


class HypothesisTestResults(TestResultsBase):
    """
    Base class for all hypothesis test results

    Parameters
    ----------
    control : Samples
        the control samples
    variation : Samples
        the variation samples
    metric : str
        The name of the metric being measured
    delta: float
        the absolute difference between the variation and control sample means
    delta_relative: float
        the percent difference between the variation and control sample means
    inference_procedure: subclass of InferenceProcedure
        the inference procedure used to generate the hypothesis test results
    metric_name : str
        The name of the metric that is being compared
    warnings : list[str]
        A list of any warning messages accumultaed during the test
    aux : dict
        Auxillary variables used for displaying or visualizing specific types
        of tests.
    """
    def __init__(self, control, variation,
                 delta, delta_relative, effect_size,
                 inference_procedure, warnings=[], aux={}):

        self.control = control
        self.variation = variation
        self.delta = delta
        self.delta_relative = delta_relative
        self.effect_size = effect_size
        self.inference_procedure = inference_procedure
        self.metric_name = None  # this gets updated by the test
        self.warnings = warnings
        self.aux = aux

    @property
    def samples_table(self):
        if not hasattr(self, "_samples_table"):
            self.render_samples_table()
        return self._samples_table

    @property
    def stats_table(self):
        if not hasattr(self, "_stats_table"):
            self.render_stats_table()
        return self._stats_table

    @property
    def table(self):
        if not hasattr(self, "_results_table"):
            self.render_tables()
        return self._results_table

    def render_samples_table(self):
        tbl = PrettyTable(header=True)
        tbl.add_column(
            "Treatment",
            [
                "Metric",
                "Observations",
                "Mean",
                "Standard Error",
                "Variance"
            ],
            align='l'
        )

        tbl.add_column(
            self.control.name,
            [
                self.metric_name,
                int(self.control.nobs),
                f"{self.control.mean:1.4f}",
                "{!r}".format(tuple([round(s, 4) for s in self.control.std_err(self.alpha)])),
                f"{self.control.var:1.4f}"
            ],
            align='l'
        )

        tbl.add_column(
            self.variation.name,
            [
                self.metric_name,
                int(self.variation.nobs),
                f"{self.variation.mean:1.4f}",
                "{!r}".format(tuple([round(s, 4) for s in self.variation.std_err(self.alpha)])),
                f"{self.variation.var:1.4f}"
            ],
            align='l'
        )
        self._samples_table = str(tbl)

    def __repr__(self):
        return self.table

    def render_tables(self):
        self.render_stats_table()
        self.render_samples_table()
        self._results_table = f"""
Observations Summary:
{self.samples_table}

Test Results:
{self.stats_table}
        """

    @property
    def _base_json(self):
        """
        Parameters that belong to all hypothesis test results
        """
        return OrderedDict([('metric', [self.metric_name]),
                            ('hypothesis', [self.hypothesis]),
                            ('model_name', [self.model_name]),
                            ('accept_hypothesis', [self.accept_hypothesis]),

                            ('control_name', [self.control.name]),
                            ('control_nobs', [self.control.nobs]),
                            ('control_mean', [self.control.mean]),
                            ('control_ci', [self.control.ci(self.alpha)]),
                            ('control_var', [self.control.var]),

                            ('variation_name', [self.variation.name]),
                            ('variation_nobs', [self.variation.nobs]),
                            ('variation_mean', [self.variation.mean]),
                            ('variation_ci', [self.variation.ci(self.alpha)]),
                            ('variation_var', [self.variation.var]),

                            ('delta', [self.delta]),
                            ('delta_relative', [100 * self.delta_relative]),
                            ('effect_size', [self.effect_size]),
                            ('alpha', [self.alpha]),
                            ('segmentation', [self.segmentation]),
                            ('warnings', [process_warnings(self.warnings)])
                            ])

    def display(self):
        raise NotImplementedError

    def visualize(self, figsize=None, outfile=None, *args, **kwargs):
        raise NotImplementedError


def process_warnings(warnings):
    if isinstance(warnings, list):
        warnings = ';'.join(warnings)
    return warnings


class HypothesisTestSuite(object):
    """
    Parameters
    ----------
    tests : list[HypothesisTest]
        The tests to run
    correction_method : str
        One of the following correction methods:
            'bonferroni', 'b' : one-step Bonferroni correction
            'sidak', 's' : one-step Sidak correction
            'fdr_bh', 'bh; : Benjamini/Hochberg (non-negative)
    """
    def __init__(self, tests, correction_method='sidak'):

        if not isinstance(tests, (list, tuple)):
            raise ValueError('`tests` must be a sequence of `HypothesisTest` instances')

        for test in tests:
            if not issubclass(test.__class__, HypothesisTest):
                raise ValueError("All tests must subclass `HypothesisTest`")

        if correction_method not in set(list(CORRECTIONS.keys()) + list(CORRECTIONS.values())):
            raise ValueError(f'Correction method {correction_method} not supported')

        self.tests = tests
        self.correction_method = CORRECTIONS[correction_method] \
                                 if correction_method in CORRECTIONS \
                                 else correction_method


class HypothesisTestSuiteResults(TestResultsBase):
    """
    Store, display, visualize, and export the results of statistical test
    suite.
    """
    def __init__(self, tests, original_results, corrected_results, correction):
        self.tests = tests
        self.ntests = len(tests)
        self.original_results = original_results
        self.corrected_results = corrected_results
        self.correction = correction

    def __repr__(self):
        return f"{self.__class__.__name__}(ntests={self.ntests}, correction_method='{self.correction.method}')"

    def display(self):
        for ii, res in enumerate(self.corrected_results):
            print('-' * 60)
            print(f'Test {ii + 1} of {self.ntests}')
            print(res)

    def visualize(self, figsize=None, outfile=None, *args, **kwargs):
        for ii, res in enumerate(self.corrected_results):
            res.visualize()



#!/usr/bin/python
# -*- coding: utf-8 -*-
from abra.config import DEFAULT_ALPHA, DEFAULT_BAYESIAN_INFERENCE_METHOD
from abra.inference.inference_base import InferenceProcedure
from abra.stats import Samples
from abra.vis import Traces
from abra.inference.bayesian import get_stan_model, get_model_datatype, get_model_data
from abra.inference.bayesian.results import BayesianTestResults
import numpy as np


SAMPLING_DEFAULT_INFERENCE_KWARGS = dict(iter=2000, chains=4, n_jobs=-1, seed=1,
                                         control=dict(stepsize=0.01,
                                                      adapt_delta=0.99))
VB_DEFAULT_INFERENCE_KWARGS = dict(iter=10000, seed=1)


class BayesianDelta(InferenceProcedure):
    """
    Runs Bayesian inference procedure to test for the difference between two
    samples. Uses Stan to run inference proceduress.
    Parameters
    ----------
    model_name: str
        The name of the model to use:
            "gaussian", "g"       : Gaussian model (symetric, continuous)
            "student-t", "t"      : Student's t model (symetric, continuous)
            "binomial", "b"       : Binomial model (binary)
            "beta-binomial", "bb" : Beta-Binomial model (binary)
            "bernoulli", "bn"     : Bernoulli model (binary)
            "poisson", "p"        : Poisson model (counts)
            "gamma-poisson", "gp" : Gamma-Poisson model (counts)
    model_params: dict
        Arguments specific for initializing each type of Bayesian model.
        See bayesian inference README for details on specifications for each model.
    hypothesis: str
        The alternative hypothesis, or how we think the variation is related to
        the control.
    """

    def __init__(self,
                 model_name,
                 model_params={},
                 hypothesis='larger',
                 *args, **kwargs):

        super(BayesianDelta, self).__init__(*args, **kwargs)
        self.model_name = model_name
        self.data_type = get_model_datatype(model_name)
        self.hypothesis = hypothesis
        self.model_params = model_params
        self.bm = get_stan_model(model_name, **model_params)
        self.trace = None

    def ensure_dtypes(self, data, name=None):
        if len(data) == 0:
            raise ValueError("No observations provided for {!r} treatment".format(name))
        return np.array(data, dtype=self.data_type)

    def inference_inputs(self, control, variation):
        """
        Processes the control and variations data for the particular model
        """
        control_observations = self.ensure_dtypes(control.data, 'control')
        variation_observations = self.ensure_dtypes(variation.data, 'variation')
        return get_model_data(control_observations, variation_observations, self.model_name, self.model_params)

    @property
    def summary(self):
        return self.traces.summary

    def plot_trace(self, variable, *args, **kwargs):
        return self.traces.plot(variable, *args, **kwargs)

    def run(self, control_samples, variation_samples,
            alpha=DEFAULT_ALPHA, inference_kwargs=None):
        """
        Run the inference procedure over the samples using a particular
        inference method.
        control_samples: instance of Samples
            the samples for the control condition
        variation_smaples: instance of Samples
            the samples for the varitation condition
        alpha: float
            One minus the credible mass under the posterior, used to calculate
            the credible interval. Roughly analogous to the alpha in frequentist
            tests.
        inference_kwargs : dict
            Additional kwargs that used by the Stan sampling or variational
            inference methods. For details, see:
              - http://mc-stan.org/rstan/reference/stanmodel-method-sampling.html
              - http://mc-stan.org/rstan/reference/stanmodel-method-vb.html
        """
        if isinstance(control_samples, (list, np.ndarray)):
            control_samples = Samples(control_samples)

        if isinstance(variation_samples, (list, np.ndarray)):
            variation_samples = Samples(variation_samples)

        self.alpha = alpha
        self.credible_mass = 1 - alpha
        self.control = control_samples
        self.variation = variation_samples
        inference_data, hyperparameters = self.inference_inputs(control_samples, variation_samples)
        self.hyperparameters = {k:v for k, v in list(hyperparameters.items()) if k!='recompile'}

        # check for `inference_method` in inference_kwargs
        if inference_kwargs is not None:
            if 'inference_method' in inference_kwargs:
                inference_method = inference_kwargs['inference_method']
                del inference_kwargs['inference_method']
            else:
                inference_method = DEFAULT_BAYESIAN_INFERENCE_METHOD
        else:
            inference_method = DEFAULT_BAYESIAN_INFERENCE_METHOD

        self.inference_method = inference_method

        # run inference
        if inference_method[0] == "s":  # Gibs sampling
            if inference_kwargs is None:
                inference_kwargs = SAMPLING_DEFAULT_INFERENCE_KWARGS
            inference_results = self.bm.sampling(data=inference_data, **inference_kwargs)

        elif inference_method[0] == "v":  # variational bayes
            if inference_kwargs is None:
                inference_kwargs = VB_DEFAULT_INFERENCE_KWARGS
            inference_results = self.bm.vb(data=inference_data, **inference_kwargs)
        else:
            raise ValueError('Unknown inference procedure {!r}'.format(inference_method))

        self.traces = Traces(self.extract_traces(inference_results))

        return self.traces

    def extract_traces(self, inference_results):
        if issubclass(inference_results.__class__, dict):
            traces = {}
            for ii in range(len(inference_results['sampler_param_names'])):
                param_name = inference_results['sampler_param_names'][ii]
                param_value = np.array(inference_results['sampler_params'][ii])
                traces[param_name] = param_value
        else:
            traces = inference_results.extract()
        return traces

    @property
    def stats(self):
        summary = self.summary
        delta = summary.loc['delta', 'mean']
        delta_relative = summary.loc['delta_relative', 'mean']
        effect_size = summary.loc['effect_size', 'mean']
        prob_greater = self.traces.delta.prob_greater_than(0.)
        return delta, delta_relative, effect_size, prob_greater

    def hdi(self, alpha, variable='delta'):
        """
        Calculate the highest density interval for the delta posterior.
        """
        values = getattr(self.traces, variable).hdi(alpha)
        percentiles = (alpha, 1 - alpha)
        return (values, percentiles)

    @property
    def hypothesis_text(self):
        variation_name = self.variation.name
        variation_name = variation_name if variation_name else 'variation'
        if self.hypothesis == 'larger':
            return "{} is larger".format(variation_name)
        elif self.hypothesis == 'smaller':
            return "{} is smaller".format(variation_name)
        else:
            return "{} != {}".format(variation_name, self.control.name)

    def make_results(self):
        """
        Package up inference results
        """
        delta, delta_relative, effect_size, prob_greater = self.stats
        hdi = self.hdi(self.alpha)
        hdi_relative = self.hdi(self.alpha, 'delta_relative')
        return BayesianTestResults(control=self.control,
                                   variation=self.variation,
                                   delta=delta,
                                   delta_relative=delta_relative,
                                   effect_size=effect_size,
                                   alpha=self.alpha,
                                   traces=self.traces,
                                   hdi=hdi,
                                   hdi_relative=hdi_relative,
                                   prob_greater=prob_greater,
                                   model_name=self.model_name,
                                   hypothesis=self.hypothesis_text,
                                   inference_method=self.inference_method,
                                   data_type=self.data_type,
                                   inference_procedure=self,
                                   hyperparameters=self.hyperparameters
                                   )



#!/usr/bin/python
# -*- coding: utf-8 -*-

from abra.config import DEFAULT_ALPHA, MIN_OBS_FOR_Z
from abra.stats import Samples, MeanComparison
from abra.inference.frequentist.results import FrequentistTestResults
from abra.inference import FrequentistProcedure
from numpy import ndarray


class MeansDelta(FrequentistProcedure):
    """
    Runs frequentist inference procedure to test for the difference in two sample
    means. Assumes normality for large sample sizes, or t-distribution for small
    sample sizes.

    Parameters
    ----------
    hypothesis : str
        the althernative hypothesis:
        - 'larger': one-tailed test, assume variation is larger than null
        - 'smaller': one-tailed test, assume variation is smaller than null
        - 'unequal': two-tailed test, variaion mean is different than null
    var_assumptions : str
        whether to use pooled or unequal variance assumptions
        - 'pooled': assume the same variance
        - 'unequal': use Smith-Satterthwait dof when calculating t-stat
    """
    def __init__(self, var_assumptions='unequal', *args, **kwargs):
        super(MeansDelta, self).__init__(*args, **kwargs)
        self.var_assumptions = var_assumptions

    def run(self, control_samples, variation_samples,
            alpha=DEFAULT_ALPHA, inference_kwargs=None):
        """
        Run the inference procedure over the samples with a selected alpha
        value.

        alpha : float in [0, 1]
            the assumed Type I error rate
        """
        if isinstance(control_samples, (list, ndarray)):
            control_samples = Samples(control_samples)

        if isinstance(variation_samples, (list, ndarray)):
            variation_samples = Samples(variation_samples)

        self.alpha = alpha

        nobs = min(control_samples.nobs, variation_samples.nobs)
        test_statistic = 'z' if nobs > MIN_OBS_FOR_Z else 't'
        self.comparison = MeanComparison(
            samples_a=variation_samples,
            samples_b=control_samples,
            alpha=self.alpha,
            test_statistic=test_statistic,
            hypothesis=self.hypothesis
        )

    @property
    def stats(self):
        f_stats = getattr(self.comparison, "{}test_ind".format(self.test_statistic))
        return f_stats(alternative=self.hypothesis_sm)

    @property
    def ci(self):
        """
        Calculate confidence interval percentiles and values.
        """
        f_ci = getattr(self.comparison, "{}confint_diff".format(self.test_statistic))
        ci_vals = f_ci(self.alpha, self.hypothesis_sm, self.var_assumptions)

        return [ci_vals, self.ci_percents]

    def make_results(self):
        """
        Package up inference results
        """
        stats = self.stats
        statistic_value = stats[0]
        p_value = stats[1]

        accept_hypothesis = self.accept_hypothesis(statistic_value)
        df = stats[2] if self.test_statistic == 't' else None

        return FrequentistTestResults(
            control=self.comparison.d2,
            variation=self.comparison.d1,
            delta=self.comparison.delta,
            delta_relative=self.comparison.delta_relative,
            effect_size=self.comparison.effect_size,
            alpha=self.comparison.alpha,
            power=self.comparison.power,
            confidence_interval=self.ci,
            test_statistic=self.test_statistic,
            statistic_value=statistic_value,
            p_value=p_value,
            df=df,
            hypothesis=self.hypothesis_text,
            accept_hypothesis=accept_hypothesis,
            inference_procedure=self,
            warnings=self.comparison.warnings
        )



import os
import sys
import logging
from pystan import StanModel
import pickle as pickle
from abra.config import STAN_MODEL_CACHE


HERE = os.path.dirname(os.path.abspath(__file__))
STAN_MODEL_HOME = os.path.join(HERE, 'models')

DEFAULT_BETA_ALPHA = 1.
DEFAULT_BETA_BETA = 1.
DEFAULT_MAX_STD = 30.
DEFAULT_PRECISION = 30.


if not os.path.isdir(STAN_MODEL_CACHE):
    os.makedirs(STAN_MODEL_CACHE)


def get_stan_model_code(model_name, **model_args):
    clean_model_name = model_name.replace("-", "_").replace(" ", "_")

    # Continuous models
    if clean_model_name in ('normal', 'gaussian', 'g'):
        from .models.continuous import gaussian as model_code_function
    elif clean_model_name in ('student_t', 't', 'exp_student_t', 'est'):
        from .models.continuous import exp_student_t as model_code_function

    # Binary / proportions models
    elif clean_model_name in ('binomial', 'b', 'beta_binomial', 'bb'):
        from .models.binary import beta_binomial as model_code_function
    elif clean_model_name in ('bernoulli', 'bn'):
        from .models.binary import bernoulli as model_code_function

    # Counts / rate models
    elif clean_model_name in ('poisson', 'p', 'gamma_poisson', 'gp'):
        from .models.counts import gamma_poisson as model_code_function

    return model_code_function()


def get_stan_model(model_name, recompile=False, **model_params):
    """
    Compile a Stan probabilistic model, or load pre-compiled model from cache,
    if available.

    Parameters
    ----------
    model_name : str
        Then name of the model used
            "bernoulli", "b": Bernoulli likelihood flat prior on probabilities
            "beta-binomial", "bb": Binomial likelihood and Beta prior
    recompile : boolean
        If set to to True, always recompile the model, otherwise try to use
        the cached pickle of the model.
    """

    python_version = 'python{0[0]}.{0[1]}'.format(sys.version_info)

    filename = "-".join([_f for _f in [model_name, python_version] if _f])
    compiled_model_file = os.path.join(STAN_MODEL_CACHE, filename + ".pickle")

    if os.path.isfile(compiled_model_file) and not recompile:
        with open(compiled_model_file, 'rb') as m:
            model = pickle.load(m)
    else:
        model_code = get_stan_model_code(model_name)
        model = StanModel(model_code=model_code)
        logging.info('Saving model to {}'.format(compiled_model_file))
        with open(compiled_model_file, 'wb') as f:
            pickle.dump(model, f)

    return model


def get_model_datatype(model_name):
    clean_model_name = model_name.replace("-", "_").replace(" ", "_")
    if clean_model_name in ('gaussian', 'g',
                            'student_t', 'st',
                            'exp_student_t', 'est'):
        return float
    elif clean_model_name in ('bernoulli', 'bn',
                              'beta_binomial', 'bb',
                              'binomial', 'b',
                              'poisson', 'p',
                              'gamma_poisson', 'gp'):
        return int
    raise ValueError('Unknown model name {!r}'.format(model_name))


def get_model_data(control_observations, variation_observations, model_name, model_params={}):
    """
    Performs any model-specific preprocessing for observations or model
    hyperparameters. We deal with hyperprams as model data here so that there
    is no need to re-compile Stan models for different model configurations.

    Parameters
    ----------
    control_observations : ndarray
        The observattions vector for the control treatment
    variation_observations : ndarray
        The observattions vector for the variation treatment
    model_name : str
        The name of one of the supported Bayesian Stan models
    model_params : dict
        Any additional model keyword arguments (e.g. hyperparameters) to be
        included as input data to the Stan model. For details on model-specific
        arguments see the model definitions in the ./models/* directory

    Returns
    -------
    iput_data : dict
        Data variable for input to the PyStan model interface.
    """

    clean_model_name = model_name.replace("-", "_").replace(" ", "_")

    # # Deal with input data
    # Models that use raw sequence inputs
    if clean_model_name in ('bernoulli',
                            'gaussian',
                            'student_t',
                            'exp_student_t',
                            'gamma_poisson'):
        input_data = dict(
            n_control=int(len(control_observations)),
            n_variation=int(len(variation_observations)),
            control=control_observations,
            variation=variation_observations
        )
    # Models that use summary statistics inputs
    elif clean_model_name in ('beta_binomial', 'bb', 'binomial'):
        input_data = dict(
            n_control=int(len(control_observations)),
            n_variation=int(len(variation_observations)),
            s_control=int(sum(control_observations)),
            s_variation=int(sum(variation_observations)),
        )

    # # Deal with model-specific hyperparameters passed in `model_params`
    hyperparameters = {}
    if clean_model_name in ('bernoulli', 'beta_binomial', 'binomial', 'gamma_poisson'):
        alpha = model_params.get('alpha', DEFAULT_BETA_ALPHA)
        beta = model_params.get('beta', DEFAULT_BETA_BETA)
        hyperparameters.update(dict(alpha_=float(alpha), beta_=float(beta)))

    elif clean_model_name in ('gaussian', 'student_t', 'exp_student_t'):
        std_max = model_params.get('std_max',  DEFAULT_MAX_STD)
        hyperparameters.update(dict(std_max=std_max))

        if clean_model_name in ('student_t', 'exp_student_t'):
            precision = model_params.get('precision', DEFAULT_PRECISION)
            hyperparameters.update(dict(precision=precision))

    input_data.update(hyperparameters)

    return input_data, hyperparameters



#!/usr/bin/python
# -*- coding: utf-8 -*-
from abra.config import search_config
from pandas import DataFrame


class DatasetException(Exception):
    pass


class Dataset(object):
    """
    Interface between raw data, global configuration, and experiment
    """
    def __init__(self, df, treatment=None, measures=None, attributes=None, meta=None):
        treatment = treatment if treatment else search_config(df, "experiment", "treatment")[0]   # always first
        if treatment is None:
            raise DatasetException(f'treatment column {treatment} not in dataframe')
        else:
            self.treatment = treatment

        if isinstance(measures, str):
            measures = [measures]

        if isinstance(meta, str):
            meta = [meta]

        self.meta = meta if meta else []
        self.measures = measures if measures \
            else search_config(df, "experiment", "measures")
        self.attributes = attributes if attributes \
            else search_config(df, "experiment", "attributes")

        all_columns = [self.treatment] + self.measures + self.attributes + self.meta
        self.data = df[all_columns]
        self.columns = set(all_columns)

    def __repr__(self):
        return f"Dataset(measures={self.measures}, attributes={self.attributes})"

    @property
    def cohorts(self):
        """
        Return a list of cohorts defined by the experiment treatment.
        """
        if not hasattr(self, '_cohorts'):
            self._cohorts = sorted(self.data[self.treatment].unique().tolist())
        return self._cohorts

    def segments(self, attribute):
        """
        Return a list of tuples containing (treatment, segment) pairs, for a
        given segmentation attribute.
        """
        return self.data.groupby([self.treatment, attribute]).sum().index.tolist()

    @property
    def cohort_measures(self):
        """
        Return metric samples for each cohort.
        """
        measures = {}
        for cohort in self.cohorts:
            measures[cohort] = {}
            for metric in self.measures:
                mask = self.data[self.treatment] == cohort
                measures[cohort][metric] = self.data[mask][metric].values

        return DataFrame(measures).T

    def segment_samples(self, attribute):
        """
        Return samples from each segment (treatment-attribute pair).
        """
        measures = {}
        for segment in self.segments(attribute):
            measures[segment] = {}
            for metric in self.measures:
                mask = (self.data[self.treatment] == segment[0]) \
                    & (self.data[attribute] == segment[1])
                measures[segment][metric] = self.data[mask][metric].values
        return DataFrame(measures).T
